# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import Deque, List, Optional, Tuple, Type, Union

import numpy as np
import torch

from nncf.common.tensor import TensorElementsType
from nncf.common.tensor_statistics.collectors import NNCFCollectorTensorProcessor
from nncf.common.tensor_statistics.collectors import NNCFTensor
from nncf.experimental.common.tensor_statistics.collectors import AbsMaxReducer
from nncf.experimental.common.tensor_statistics.collectors import AbsQuantileReducer
from nncf.experimental.common.tensor_statistics.collectors import AggregatorBase
from nncf.experimental.common.tensor_statistics.collectors import BatchMeanReducer
from nncf.experimental.common.tensor_statistics.collectors import MaxAggregator
from nncf.experimental.common.tensor_statistics.collectors import MaxReducer
from nncf.experimental.common.tensor_statistics.collectors import MeanAggregator
from nncf.experimental.common.tensor_statistics.collectors import MeanPerChReducer
from nncf.experimental.common.tensor_statistics.collectors import MeanReducer
from nncf.experimental.common.tensor_statistics.collectors import MedianAbsoluteDeviationAggregator
from nncf.experimental.common.tensor_statistics.collectors import MinAggregator
from nncf.experimental.common.tensor_statistics.collectors import MinReducer
from nncf.experimental.common.tensor_statistics.collectors import NoopAggregator
from nncf.experimental.common.tensor_statistics.collectors import NoopReducer
from nncf.experimental.common.tensor_statistics.collectors import PercentileAggregator
from nncf.experimental.common.tensor_statistics.collectors import QuantileReducer
from nncf.experimental.common.tensor_statistics.collectors import RawReducer
from nncf.experimental.common.tensor_statistics.collectors import ShapeAggregator
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.quantization.advanced_parameters import StatisticsType
from nncf.torch.tensor import PTNNCFTensor
from nncf.torch.tensor_statistics.statistics import PTMeanTensorStatistic
from nncf.torch.tensor_statistics.statistics import PTMedianMADTensorStatistic
from nncf.torch.tensor_statistics.statistics import PTMinMaxTensorStatistic
from nncf.torch.tensor_statistics.statistics import PTPercentileTensorStatistic
from nncf.torch.tensor_statistics.statistics import PTRawTensorStatistic


class PTNNCFCollectorTensorProcessor(NNCFCollectorTensorProcessor):
    """
    A realization of the processing methods for PTNNCFTensors.
    """

    @staticmethod
    def reduce_min(x: NNCFTensor, axis: Union[int, Tuple[int, ...], List[int]], keepdims: bool = False) -> NNCFTensor:
        return PTNNCFTensor(torch.amin(x.tensor, dim=axis, keepdim=keepdims))

    @staticmethod
    def reduce_max(x: NNCFTensor, axis: Union[int, Tuple[int, ...], List[int]], keepdims: bool = False) -> NNCFTensor:
        return PTNNCFTensor(torch.amax(x.tensor, dim=axis, keepdim=keepdims))

    @staticmethod
    def abs(x: NNCFTensor) -> NNCFTensor:
        return PTNNCFTensor(torch.abs(x.tensor))

    @classmethod
    def min(cls, *args) -> NNCFTensor:
        stacked = cls.stack(args)
        return cls.reduce_min(stacked, axis=0, keepdims=False)

    @classmethod
    def max(cls, *args) -> NNCFTensor:
        stacked = cls.stack(args)
        return cls.reduce_max(stacked, axis=0, keepdims=False)

    @staticmethod
    def mean(x: NNCFTensor, axis: Union[int, Tuple[int, ...], List[int]], keepdims=False) -> NNCFTensor:
        return PTNNCFTensor(x.tensor.mean(dim=axis, keepdim=keepdims))

    @staticmethod
    def median(x: NNCFTensor, axis: Union[int, Tuple[int, ...], List[int]], keepdims=False) -> NNCFTensor:
        # See https://github.com/pytorch/pytorch/issues/61582
        if not isinstance(axis, int):
            device = x.tensor.device
            result = torch.tensor(np.median(x.tensor.detach().cpu().numpy(), axis=axis, keepdims=keepdims))
            return PTNNCFTensor(result.type(x.tensor.dtype).to(device))
        return PTNNCFCollectorTensorProcessor.quantile(x, quantile=[0.5], axis=axis, keepdims=keepdims)[0]

    @classmethod
    def masked_mean(
        cls, x: NNCFTensor, axis: Union[int, Tuple[int, ...]], mask: NNCFTensor, keepdims: bool = False
    ) -> NNCFTensor:
        if mask is None:
            return cls.mean(x, axis=axis, keepdims=keepdims)
        device = x.tensor.device
        masked_x = np.ma.array(x.tensor.detach().cpu().numpy(), mask=mask.tensor.detach().cpu().numpy())
        result = np.ma.mean(masked_x, axis=axis, keepdims=keepdims).astype(masked_x.dtype)
        if isinstance(result, np.ma.MaskedArray):
            result = result.data
        return PTNNCFTensor(torch.tensor(result).to(device=device))

    @classmethod
    def masked_median(
        cls, x: NNCFTensor, axis: Union[int, Tuple[int, ...]], mask: NNCFTensor, keepdims: bool = False
    ) -> NNCFTensor:
        # Implemented in numy as torch.masked.median is not implemented yet
        if mask is None:
            return cls.median(x, axis=axis, keepdims=keepdims)
        device = x.tensor.device
        masked_x = np.ma.array(x.tensor.detach().cpu().numpy(), mask=mask.tensor.detach().cpu().numpy())
        result = np.ma.median(masked_x, axis=axis, keepdims=keepdims).astype(masked_x.dtype)
        if isinstance(result, np.ma.MaskedArray):
            result = result.data
        return PTNNCFTensor(torch.tensor(result).to(device=device))

    @staticmethod
    def mean_per_channel(x: NNCFTensor, axis: int) -> NNCFTensor:
        if len(x.shape) < 3:
            return PTNNCFTensor(torch.mean(x.tensor, axis=0))
        x = torch.moveaxis(x.tensor, axis, 1)
        t = x.reshape(x.shape[0], x.shape[1], -1)
        return PTNNCFTensor(torch.mean(t, axis=(0, 2)))

    @staticmethod
    def batch_mean(x: NNCFTensor) -> NNCFTensor:
        return PTNNCFTensor(torch.mean(x.tensor, axis=0, keepdims=True))

    @staticmethod
    def transpose(x: NNCFTensor, axes: Tuple[int, ...]) -> NNCFTensor:
        return PTNNCFTensor(torch.permute(x.tensor, axes))

    @staticmethod
    def reshape(x: NNCFTensor, shape: Tuple[int, ...]) -> NNCFTensor:
        return PTNNCFTensor(torch.reshape(x.tensor, shape))

    @staticmethod
    def cat(x: List[NNCFTensor], axis: int) -> NNCFTensor:
        x = [t.tensor for t in x]
        return PTNNCFTensor(torch.cat(x, axis))

    @staticmethod
    def logical_or(input_: NNCFTensor, other: NNCFTensor) -> NNCFTensor:
        return PTNNCFTensor(torch.logical_or(input_.tensor, other.tensor))

    @staticmethod
    def less(input_: NNCFTensor, other: NNCFTensor) -> NNCFTensor:
        return PTNNCFTensor(input_.tensor < other.tensor)

    @staticmethod
    def stack(x: Union[List[NNCFTensor], Deque[NNCFTensor]], axis: int = 0) -> NNCFTensor:
        x = [t.tensor for t in x]
        return PTNNCFTensor(torch.stack(x, dim=axis))

    @staticmethod
    def unstack(x: NNCFTensor, axis: int = 0) -> List[NNCFTensor]:
        tensor = x.tensor
        if list(tensor.shape) == []:
            tensor = tensor.unsqueeze(0)
        tensor_list = torch.unbind(tensor, dim=axis)
        return [PTNNCFTensor(t) for t in tensor_list]

    @staticmethod
    def squeeze(x: NNCFTensor, dim: Optional[Union[int, Tuple[int, ...]]] = None) -> NNCFTensor:
        return PTNNCFTensor(torch.squeeze(x.tensor, dim=dim))

    @staticmethod
    def sum(tensor: NNCFTensor) -> TensorElementsType:
        return torch.sum(tensor.tensor).item()

    @staticmethod
    def quantile(
        tensor: NNCFTensor,
        quantile: Union[float, List[float], np.ndarray],
        axis: Union[int, Tuple[int, ...], List[int]],
        keepdims: bool = False,
    ) -> List[NNCFTensor]:
        device = tensor.device
        # See https://github.com/pytorch/pytorch/issues/61582
        # https://github.com/pytorch/pytorch/issues/64947
        if len(tensor.tensor) <= 16_000_000 and isinstance(axis, int):
            result = torch.quantile(
                tensor.tensor,
                torch.tensor(quantile, dtype=tensor.tensor.dtype, device=tensor.tensor.device),
                axis,
                keepdims,
            )
        else:
            result = torch.tensor(
                np.quantile(tensor.tensor.detach().cpu().numpy(), q=quantile, axis=axis, keepdims=keepdims)
            )
        result = result.type(tensor.tensor.dtype).to(device)
        return [PTNNCFTensor(x) for x in result]

    @classmethod
    def percentile(
        cls,
        tensor: NNCFTensor,
        percentile: Union[float, List[float], np.ndarray],
        axis: Union[int, Tuple[int, ...], List[int]],
        keepdims: bool = False,
    ) -> List[TensorElementsType]:
        quantile = np.true_divide(percentile, 100)
        return cls.quantile(tensor, quantile=quantile, axis=axis, keepdims=keepdims)

    @staticmethod
    def sub(a: NNCFTensor, b: NNCFTensor) -> NNCFTensor:
        return NNCFTensor(a.tensor - b.tensor)

    @staticmethod
    def zero_elements(x: NNCFTensor) -> NNCFTensor:
        pt_tensor = x.tensor
        eps = torch.finfo(pt_tensor.dtype).eps
        return NNCFTensor(pt_tensor.abs() < eps)


class PTReducerMixIn:
    def _get_processor(self):
        return PTNNCFCollectorTensorProcessor

    def get_inplace_fn(self):
        return None

    def get_output_names(self, target_node_name: str, port_id: int) -> List[str]:
        return []


class PTMinReducer(PTReducerMixIn, MinReducer):
    pass


class PTMaxReducer(PTReducerMixIn, MaxReducer):
    pass


class PTAbsMaxReducer(PTReducerMixIn, AbsMaxReducer):
    pass


class PTMeanReducer(PTReducerMixIn, MeanReducer):
    pass


class PTQuantileReducer(PTReducerMixIn, QuantileReducer):
    pass


class PTAbsQuantileReducer(PTReducerMixIn, AbsQuantileReducer):
    pass


class PTBatchMeanReducer(PTReducerMixIn, BatchMeanReducer):
    pass


class PTMeanPerChanelReducer(PTReducerMixIn, MeanPerChReducer):
    pass


def _reshape_all(targets: Tuple[torch.Tensor, ...], target_shape: Tuple[int, ...]):
    return map(lambda stat: torch.reshape(stat, target_shape), targets)


def _get_wrapped_min_max_tensor_statistic(target_shape: Tuple[int, ...]) -> Type[PTMinMaxTensorStatistic]:
    """
    Returns PTMinMaxTensorStatistic type but all statistics are reshaped to target_shape.

    :param target_shape: Target shape of the tensor statistic
    :return: PTMinMaxTensorStatistic type but all statistics are reshaped to target_shape.
    """

    class WrappedPTMinMaxTensorStatistic(PTMinMaxTensorStatistic):
        def __init__(self, min_values, max_values):
            min_values, max_values = _reshape_all((min_values, max_values), target_shape)
            super().__init__(min_values, max_values)

    return WrappedPTMinMaxTensorStatistic


def _get_wrapped_percentile_tensor_statistic(target_shape: Tuple[int, ...]) -> Type[PTPercentileTensorStatistic]:
    """
    Returns PTPercentileTensorStatistic type but all statistics are reshaped to target_shape.

    :param target_shape: Target shape of the tensor statistic
    :return: PTPercentileTensorStatistic type but all statistics are reshaped to target_shape.
    """

    class WrappedPTPercentileTensorStatistic(PTPercentileTensorStatistic):
        def __init__(self, percentile_vs_values_dict):
            reshaped_percentiles = {}
            for k, v in percentile_vs_values_dict.items():
                reshaped_percentiles[k] = torch.reshape(v, target_shape)
            super().__init__(reshaped_percentiles)

    return WrappedPTPercentileTensorStatistic


def get_min_max_statistic_collector(
    use_abs_max: bool,
    reduction_axes: Tuple[int, ...],
    aggregation_axes: Tuple[int, ...],
    scale_shape: Tuple[int, ...],
    num_samples: int,
) -> TensorCollector:
    """
    Min max statistic collector builder.

    :param use_abs_max: Whether to use abs max reducer or max reducer.
    :param reduction_axes: Axes to use in reduction functions.
    :param aggregation_axes: Axes to use in aggregation functions.
    :param scale_shape: Target shape for collected statistics.
    :param num_samples: Maximum number of samples to collect.
    :return: Min max statistic collector.
    """

    tensor_collector = TensorCollector(_get_wrapped_min_max_tensor_statistic(target_shape=scale_shape))

    aggregator_kwargs = {
        "tensor_processor": PTNNCFCollectorTensorProcessor,
        "num_samples": num_samples,
        "aggregation_axes": aggregation_axes,
    }
    min_reducer = PTMinReducer(reduction_axes)
    min_aggregator = MinAggregator(**aggregator_kwargs)
    tensor_collector.register_statistic_branch(PTMinMaxTensorStatistic.MIN_STAT, min_reducer, min_aggregator)

    max_reducer_cls = PTAbsMaxReducer if use_abs_max else PTMaxReducer
    max_reducer = max_reducer_cls(reduction_axes)
    max_aggregator = MaxAggregator(**aggregator_kwargs)
    tensor_collector.register_statistic_branch(PTMinMaxTensorStatistic.MAX_STAT, max_reducer, max_aggregator)
    return tensor_collector


def get_mixed_min_max_statistic_collector(
    use_abs_max: bool,
    reduction_axes: Tuple[int, ...],
    aggregation_axes: Tuple[int, ...],
    scale_shape: Tuple[int, ...],
    use_means_of_mins: bool,
    use_means_of_maxs: bool,
    num_samples: int = None,
    window_size: Optional[int] = None,
) -> TensorCollector:
    """
    Mixed min max statistic collector builder.

    :param use_abs_max: Whether to use abs max reducer or max reducer.
    :param reduction_axes: Axes to use in reduction functions.
    :param aggregation_axes: Axes to use in aggregation functions.
    :param scale_shape: Target shape for collected statistics.
    :param use_means_of_mins: Whether to use mean or min aggregator for minimum statistic branch.
    :param use_means_of_maxs: Whether to use mean or max aggregator for maximum statistic branch.
    :param num_samples: Maximum number of samples to collect.
    :param window_size: Number of samples from the end of the list of collected samples to aggregate.
        Aggregates all available collected statistics in case parameter is None.
    :return: Mixed min max statistic collector.
    """
    tensor_collector = TensorCollector(_get_wrapped_min_max_tensor_statistic(target_shape=scale_shape))
    min_reducer = PTMinReducer(reduction_axes)

    kwargs = {
        "tensor_processor": PTNNCFCollectorTensorProcessor,
        "num_samples": num_samples,
        "aggregation_axes": aggregation_axes,
        "window_size": window_size,
    }
    min_aggregator_cls = MeanAggregator if use_means_of_mins else MinAggregator
    min_aggregator = min_aggregator_cls(**kwargs)
    tensor_collector.register_statistic_branch(PTMinMaxTensorStatistic.MIN_STAT, min_reducer, min_aggregator)

    max_reducer_cls = PTAbsMaxReducer if use_abs_max else PTMaxReducer
    max_reducer = max_reducer_cls(reduction_axes)
    max_aggregator_cls = MeanAggregator if use_means_of_maxs else MaxAggregator
    max_aggregator = max_aggregator_cls(**kwargs)
    tensor_collector.register_statistic_branch(PTMinMaxTensorStatistic.MAX_STAT, max_reducer, max_aggregator)

    return tensor_collector


def get_median_mad_statistic_collector(
    reduction_axes: Tuple[int, ...],
    aggregation_axes: Tuple[int, ...],
    scale_shape: Tuple[int, ...],
    num_samples: int,
    window_size: Optional[int] = None,
) -> TensorCollector:
    """
    Median Absolute Deviation statistic collector builder.

    :param reduction_axes: Axes to use in reduction functions.
    :param aggregation_axes: Axes to use in aggregation functions.
    :param scale_shape: Target shape for collected statistics.
    :param num_samples: Maximum number of samples to collect.
    :param window_size: Number of samples from the end of the list of collected samples to aggregate.
        Aggregates all available collected statistics in case parameter is None.
    :return: Median Absolute Deviation statistic collector.

    """

    class WrappedPTMedianMADTensorStatistic(PTMedianMADTensorStatistic):
        def __init__(self, median_values, mad_values):
            median_values, mad_values = _reshape_all((median_values, mad_values), scale_shape)
            super().__init__(median_values, mad_values)

    return _get_collection_without_reduction(
        MedianAbsoluteDeviationAggregator,
        WrappedPTMedianMADTensorStatistic,
        reduction_axes=reduction_axes,
        aggregation_axes=aggregation_axes,
        num_samples=num_samples,
        window_size=window_size,
    )


def get_percentile_tensor_collector(
    percentiles_to_collect: Tuple[int, ...],
    reduction_axes: Tuple[int, ...],
    aggregation_axes: Tuple[int, ...],
    scale_shape: Tuple[int, ...],
    num_samples: int,
    window_size: Optional[int] = None,
) -> TensorCollector:
    """
    Percentile statistic collector builder.

    :param percentiles_to_collect: Percetiles to use on aggregation phase.
    :param reduction_axes: Axes to use in reduction functions.
    :param aggregation_axes: Axes to use in aggregation functions.
    :param scale_shape: Target shape for collected statistics.
    :param num_samples: Maximum number of samples to collect.
    :param window_size: Number of samples from the end of the list of collected samples to aggregate.
        Aggregates all available collected statistics in case parameter is None.
    :return: Percentile statistic collector.
    """
    return _get_collection_without_reduction(
        partial(PercentileAggregator, percentiles_to_collect=percentiles_to_collect),
        _get_wrapped_percentile_tensor_statistic(target_shape=scale_shape),
        reduction_axes=reduction_axes,
        aggregation_axes=aggregation_axes,
        num_samples=num_samples,
        window_size=window_size,
    )


def _get_collection_without_reduction(
    aggregator_cls: AggregatorBase,
    statistic_cls: AggregatorBase,
    reduction_axes: Tuple[int, ...],
    aggregation_axes: Tuple[int, ...],
    num_samples: int,
    window_size: Optional[int] = None,
) -> TensorCollector:
    """
    Helper function to build a tensor collector which is reducing statistics exclusively during aggregation phase.

    :param aggregator_cls: Aggregator class to build the tensor collector.
    :param aggregator_cls: Statistic class to build the tensor collector.
    :param reduction_axes: Axes to use in reduction functions.
    :param aggregation_axes: Axes to use in aggregation functions.
    :param num_samples: Maximum number of samples to collect.
    :param window_size: Number of samples from the end of the list of collected samples to aggregate.
        Aggregates all available collected statistics in case parameter is None.
    :return: Target statistic collector.
    """
    tensor_collector = TensorCollector(statistic_cls)
    reducer = NoopReducer()
    aggregation_axes = list(set(list(aggregation_axes) + [dim + 1 for dim in reduction_axes]))
    aggregator = aggregator_cls(
        PTNNCFCollectorTensorProcessor,
        aggregation_axes=aggregation_axes,
        window_size=window_size,
        num_samples=num_samples,
    )

    tensor_collector.register_statistic_branch(
        PTMedianMADTensorStatistic.TENSOR_STATISTIC_OUTPUT_KEY, reducer, aggregator
    )
    return tensor_collector


def get_mean_percentile_statistic_collector(
    percentiles_to_collect: Tuple[int, ...],
    reduction_axes: Tuple[int, ...],
    aggregation_axes: Tuple[int, ...],
    scale_shape: Tuple[int, ...],
    num_samples: int,
    window_size: Optional[int] = None,
) -> TensorCollector:
    """
    Mean percentile statistic collector builder.

    :param percentiles_to_collect: Percetiles to use on reduction phase.
    :param reduction_axes: Axes to use in reduction functions.
    :param aggregation_axes: Axes to use in aggregation functions.
    :param scale_shape: Target shape for collected statistics.
    :param num_samples: Maximum number of samples to collect.
    :param window_size: Number of samples from the end of the list of collected samples to aggregate.
        Aggregates all available collected statistics in case parameter is None.
    :return: Mean percentile statistic collector.
    """
    tensor_collector = TensorCollector(_get_wrapped_percentile_tensor_statistic(target_shape=scale_shape))
    quantiles_to_collect = np.true_divide(percentiles_to_collect, 100)
    reducer = PTQuantileReducer(reduction_axes=reduction_axes, quantile=quantiles_to_collect)
    for output_port_id, p in enumerate(percentiles_to_collect):
        aggregator = MeanAggregator(
            PTNNCFCollectorTensorProcessor,
            aggregation_axes=aggregation_axes,
            num_samples=num_samples,
            window_size=window_size,
        )
        tensor_collector.register_statistic_branch(
            (PTPercentileTensorStatistic.PERCENTILE_VS_VALUE_DICT, p), reducer, aggregator, output_port_id
        )
    return tensor_collector


def get_mean_statistic_collector(
    num_samples: int, channel_axis: int, window_size: Optional[int] = None
) -> TensorCollector:
    """
    Mean statistic collector builder.

    :param num_samples: Maximum number of samples to collect.
    :param channel_axis: Channel axis to use during reduction phase.
    :param window_size: Number of samples from the end of the list of collected samples to aggregate.
        Aggregates all available collected statistics in case parameter is None.
    :return: Mean statistic collector.
    """
    if channel_axis == 0:
        reducer = PTBatchMeanReducer()
    else:
        reducer = PTMeanPerChanelReducer(channel_axis=channel_axis)
    noop_reducer = NoopReducer()

    kwargs = {
        "tensor_processor": PTNNCFCollectorTensorProcessor,
        "num_samples": num_samples,
        "window_size": window_size,
    }
    aggregate_mean = MeanAggregator(**kwargs)
    aggregate_shape = ShapeAggregator()

    collector = TensorCollector(PTMeanTensorStatistic)
    collector.register_statistic_branch(PTMeanTensorStatistic.MEAN_STAT, reducer, aggregate_mean)
    collector.register_statistic_branch(PTMeanTensorStatistic.SHAPE_STAT, noop_reducer, aggregate_shape)
    return collector


def get_raw_stat_collector(num_samples: Optional[int] = None) -> TensorCollector:
    """
    Raw statistic collector builder.

    :param num_samples: Maximum number of samples to collect.
    :return: Raw statistic collector.
    """
    reducer = RawReducer()
    aggregator = NoopAggregator(num_samples)

    collector = TensorCollector(PTRawTensorStatistic)
    collector.register_statistic_branch(PTRawTensorStatistic.VALUES_STATS, reducer, aggregator)
    return collector


PT_REDUCERS_MAP = {
    StatisticsType.MIN: PTMinReducer,
    StatisticsType.MAX: PTMaxReducer,
    StatisticsType.ABS_MAX: PTAbsMaxReducer,
    StatisticsType.MEAN: PTMeanReducer,
    StatisticsType.QUANTILE: PTQuantileReducer,
    StatisticsType.ABS_QUANTILE: PTAbsQuantileReducer,
}



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from nncf.api.statistics import Statistics
from nncf.common.pruning.statistics import FilterPruningStatistics
from nncf.common.quantization.statistics import QuantizationStatistics
from nncf.common.sparsity.statistics import ConstSparsityStatistics
from nncf.common.sparsity.statistics import MagnitudeSparsityStatistics
from nncf.common.sparsity.statistics import MovementSparsityStatistics
from nncf.common.sparsity.statistics import RBSparsityStatistics
from nncf.common.utils.api_marker import api


@api()
class NNCFStatistics(Statistics):
    """
    Groups statistics for all available NNCF compression algorithms.
    Statistics are present only if the algorithm has been started.
    """

    def __init__(self):
        """
        Initializes nncf statistics.
        """
        self._storage = {}

    @property
    def magnitude_sparsity(self) -> Optional[MagnitudeSparsityStatistics]:
        """
        Returns statistics of the magnitude sparsity algorithm. If statistics
        have not been collected, `None` will be returned.

        :return: Instance of the `MagnitudeSparsityStatistics` class.
        """
        return self._storage.get("magnitude_sparsity")

    @property
    def rb_sparsity(self) -> Optional[RBSparsityStatistics]:
        """
        Returns statistics of the RB-sparsity algorithm. If statistics
        have not been collected, `None` will be returned.

        :return: Instance of the `RBSparsityStatistics` class.
        """
        return self._storage.get("rb_sparsity")

    @property
    def movement_sparsity(self) -> Optional[MovementSparsityStatistics]:
        """
        Returns statistics of the movement sparsity algorithm. If statistics
        have not been collected, `None` will be returned.

        :return: Instance of the `MovementSparsityStatistics` class.
        """
        return self._storage.get("movement_sparsity")

    @property
    def const_sparsity(self) -> Optional[ConstSparsityStatistics]:
        """
        Returns statistics of the const sparsity algorithm. If statistics
        have not been collected, `None` will be returned.

        :return: Instance of the `ConstSparsityStatistics` class.
        """
        return self._storage.get("const_sparsity")

    @property
    def quantization(self) -> Optional[QuantizationStatistics]:
        """
        Returns statistics of the quantization algorithm. If statistics
        have not been collected, `None` will be returned.

        :return: Instance of the `QuantizationStatistics` class.
        """
        return self._storage.get("quantization")

    @property
    def filter_pruning(self) -> Optional[FilterPruningStatistics]:
        """
        Returns statistics of the filter pruning algorithm. If statistics
        have not been collected, `None` will be returned.

        :return: Instance of the `FilterPruningStatistics` class.
        """
        return self._storage.get("filter_pruning")

    def register(self, algorithm_name: str, stats: Statistics):
        """
        Registers statistics for the algorithm.

        :param algorithm_name: Name of the algorithm. Should be one of the following
            * magnitude_sparsity
            * rb_sparsity
            * const_sparsity
            * quantization
            * filter_pruning

        :param stats: Statistics of the algorithm.
        """

        available_algorithms = [
            "magnitude_sparsity",
            "rb_sparsity",
            "movement_sparsity",
            "const_sparsity",
            "quantization",
            "filter_pruning",
        ]
        if algorithm_name not in available_algorithms:
            raise ValueError(
                f"Can not register statistics for the algorithm. Unknown name of the algorithm: {algorithm_name}."
            )

        self._storage[algorithm_name] = stats

    def to_str(self) -> str:
        """
        Calls `to_str()` method for all registered statistics of the algorithm and returns
        a sum-up string.

        :return: A representation of the NNCF statistics as a human-readable string.
        """
        pretty_string = "\n\n".join([stats.to_str() for stats in self._storage.values()])
        return pretty_string

    def __iter__(self):
        return iter(self._storage.items())



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Base classes for NNCF PyTorch sparsity algorithm builder and controller objects.
"""
from typing import List

import torch

from nncf.api.compression import CompressionLoss
from nncf.api.compression import CompressionStage
from nncf.common.graph import NNCFNode
from nncf.common.graph import NNCFNodeName
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.logging import nncf_logger
from nncf.common.schedulers import BaseCompressionScheduler
from nncf.common.schedulers import StubCompressionScheduler
from nncf.common.sparsity.controller import SparsityController
from nncf.common.sparsity.schedulers import SparsityScheduler
from nncf.common.utils.api_marker import api
from nncf.common.utils.backend import copy_model
from nncf.torch.algo_selector import ZeroCompressionLoss
from nncf.torch.compression_method_api import PTCompressionAlgorithmBuilder
from nncf.torch.compression_method_api import PTCompressionAlgorithmController
from nncf.torch.graph.transformations.commands import PTInsertionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.graph.transformations.commands import TransformationPriority
from nncf.torch.graph.transformations.layout import PTTransformationLayout
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.sparsity.layers import BinaryMask
from nncf.torch.utils import get_model_device


class SparseModuleInfo:
    def __init__(self, module_node_name: NNCFNodeName, module: torch.nn.Module, operand):
        self.module_node_name = module_node_name
        self.module = module
        self.operand = operand


class BaseSparsityAlgoBuilder(PTCompressionAlgorithmBuilder):
    def __init__(self, config, should_init: bool = True):
        super().__init__(config, should_init)
        self._sparsified_module_info = []

    def _get_transformation_layout(self, target_model: NNCFNetwork) -> PTTransformationLayout:
        layout = PTTransformationLayout()
        commands = self._sparsify_weights(target_model)
        for command in commands:
            layout.register(command)
        return layout

    def _sparsify_weights(self, target_model: NNCFNetwork) -> List[PTInsertionCommand]:
        device = get_model_device(target_model)
        sparsified_module_nodes = target_model.nncf.get_weighted_original_graph_nodes(
            nncf_module_names=self.compressed_nncf_module_names
        )
        insertion_commands = []
        for module_node in sparsified_module_nodes:
            node_name = module_node.node_name

            if not self._should_consider_scope(node_name):
                nncf_logger.info(f"Ignored adding weight sparsifier for operation: {node_name}")
                continue

            compression_lr_multiplier = self.config.get_redefinable_global_param_value_for_algo(
                "compression_lr_multiplier", self.name
            )
            operation = self.create_weight_sparsifying_operation(module_node, compression_lr_multiplier)
            hook = operation.to(device)
            insertion_commands.append(
                PTInsertionCommand(
                    PTTargetPoint(TargetType.OPERATION_WITH_WEIGHTS, target_node_name=node_name),
                    hook,
                    TransformationPriority.SPARSIFICATION_PRIORITY,
                )
            )
            sparsified_module = target_model.nncf.get_containing_module(node_name)
            self._sparsified_module_info.append(SparseModuleInfo(node_name, sparsified_module, hook))

        return insertion_commands

    def create_weight_sparsifying_operation(self, target_module_node: NNCFNode, compression_lr_multiplier: float):
        raise NotImplementedError

    def initialize(self, model: NNCFNetwork) -> None:
        pass


@api()
class BaseSparsityAlgoController(PTCompressionAlgorithmController, SparsityController):
    """
    Base class for sparsity algorithm controllers in PT.
    """

    def __init__(self, target_model: NNCFNetwork, sparsified_module_info: List[SparseModuleInfo]):
        super().__init__(target_model)
        self._loss = ZeroCompressionLoss(get_model_device(target_model))
        self._scheduler = BaseCompressionScheduler()
        self.sparsified_module_info = sparsified_module_info

    @property
    def current_sparsity_level(self) -> float:
        return self._scheduler.current_sparsity_level

    @property
    def loss(self) -> CompressionLoss:
        return self._loss

    @property
    def scheduler(self) -> SparsityScheduler:
        return self._scheduler

    def disable_scheduler(self):
        self._scheduler = StubCompressionScheduler()
        self._scheduler.target_level = 0.0
        self._scheduler.current_sparsity_level = 0.0

    def compression_stage(self) -> CompressionStage:
        return CompressionStage.FULLY_COMPRESSED

    def strip_model(self, model: NNCFNetwork, do_copy: bool = False) -> NNCFNetwork:
        if do_copy:
            model = copy_model(model)

        for node in model.nncf.get_original_graph().get_all_nodes():
            if node.node_type in ["nncf_model_input", "nncf_model_output"]:
                continue
            nncf_module = model.nncf.get_containing_module(node.node_name)
            if hasattr(nncf_module, "pre_ops"):
                for key in list(nncf_module.pre_ops.keys()):
                    op = nncf_module.get_pre_op(key)
                    if isinstance(op.operand, BinaryMask):
                        nncf_module.weight.data = op.operand.apply_binary_mask(nncf_module.weight.data)
                        nncf_module.remove_pre_forward_operation(key)

        return model



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from dataclasses import dataclass
from dataclasses import field
from typing import List, Optional, Set

import nncf
from nncf.common.graph.graph import NNCFGraph
from nncf.common.logging import nncf_logger
from nncf.common.utils.api_marker import api


@api(canonical_alias="nncf.Subgraph")
@dataclass
class Subgraph:
    """
    Defines the ignored subgraph as follows: A subgraph comprises all nodes along
    all simple paths in the graph from input to output nodes.

    :param inputs: Input node names.
    :type inputs: List[str]
    :param outputs: Output node names.
    :type outputs: List[str]
    """

    inputs: List[str] = field(default_factory=list)
    outputs: List[str] = field(default_factory=list)


def get_ignored_node_names_from_subgraph(graph: NNCFGraph, subgraph: Subgraph) -> List[str]:
    """
    Returns all names that should be ignored according to given subgraph.

    :param graph: Given NNCFGraph.
    :param subgraph: Given subgraph instance.
    :return: All names that should be ignored according to given subgraph.
    """
    ignored_names = set()
    for start_node_name in subgraph.inputs:
        for end_node_name in subgraph.outputs:
            for path in graph.get_all_simple_paths(start_node_name, end_node_name):
                for node_key in path:
                    node = graph.get_node_by_key(node_key)
                    ignored_names.add(node.node_name)

    return list(sorted(ignored_names))


@api(canonical_alias="nncf.IgnoredScope")
@dataclass
class IgnoredScope:
    """
    Provides an option to specify portions of model to be excluded from compression.

    The ignored scope defines model sub-graphs that should be excluded from the compression process such as
    quantization, pruning and etc.

    Example:

    ..  code-block:: python

            import nncf

            # Exclude by node name:
            node_names = ['node_1', 'node_2', 'node_3']
            ignored_scope = nncf.IgnoredScope(names=node_names)

            # Exclude using regular expressions:
            patterns = ['node_\\d']
            ignored_scope = nncf.IgnoredScope(patterns=patterns)

            # Exclude by operation type:

            # OpenVINO opset https://docs.openvino.ai/latest/openvino_docs_ops_opset.html
            operation_types = ['Multiply', 'GroupConvolution', 'Interpolate']
            ignored_scope = nncf.IgnoredScope(types=operation_types)

            # ONNX opset https://github.com/onnx/onnx/blob/main/docs/Operators.md
            operation_types = ['Mul', 'Conv', 'Resize']
            ignored_scope = nncf.IgnoredScope(types=operation_types)

    **Note:** Operation types must be specified according to the model framework.

    :param names: List of ignored node names.
    :type names: List[str]
    :param patterns: List of regular expressions that define patterns for names of ignored nodes.
    :type patterns: List[str]
    :param types: List of ignored operation types.
    :type types: List[str]
    :param subgraphs: List of ignored subgraphs.
    :type subgraphs: List[Subgraph]
    :param validate: If set to True, then a RuntimeError will be raised if any ignored scope does not match
      in the model graph.
    :type types: bool
    """

    names: List[str] = field(default_factory=list)
    patterns: List[str] = field(default_factory=list)
    types: List[str] = field(default_factory=list)
    subgraphs: List[Subgraph] = field(default_factory=list)
    validate: bool = True


def convert_ignored_scope_to_list(ignored_scope: Optional[IgnoredScope]) -> List[str]:
    """
    Convert the contents of the `IgnoredScope` class to the legacy ignored
    scope format.

    :param ignored_scope: The ignored scope.
    :return: An ignored scope in the legacy format as list.
    """
    results = []
    if ignored_scope is None:
        return results
    results.extend(ignored_scope.names)
    for p in ignored_scope.patterns:
        results.append("{re}" + p)
    if ignored_scope.types:
        raise nncf.InternalError("Legacy ignored scope format does not support operation types")
    return results


def get_ignored_node_names_from_ignored_scope(
    ignored_scope: IgnoredScope, nncf_graph: NNCFGraph, strict: bool = True
) -> Set[str]:
    """
    Returns ignored names according to ignored scope and NNCFGraph.
    If strict is True, raises RuntimeError if any ignored name is not found in the NNCFGraph or
    any ignored pattern or any ignored type match 0 nodes in the NNCFGraph.
    If strict is False, returns all possible matches.

    :param ignored_scope: Given ignored scope instance.
    :param nncf_graph: Given NNCFGraph.
    :param strict: Whether all ignored_scopes must match at least one node or not.
    :returns: NNCF node names from given NNCFGraph specified in given ignored scope.
    """
    error_msg = (
        "Refer to the original_graph.dot to discover the operations"
        "in the model currently visible to NNCF and specify the ignored/target"
        " scopes in terms of the names there."
    )

    node_names = [node.node_name for node in nncf_graph.get_all_nodes()]
    matched_by_names = []
    if ignored_scope.names:
        for ignored_node_name in ignored_scope.names:
            if ignored_node_name in node_names:
                matched_by_names.append(ignored_node_name)
        if strict and len(ignored_scope.names) != len(matched_by_names):
            skipped_names = set(ignored_scope.names) - set(matched_by_names)
            raise nncf.ValidationError(
                f"Ignored nodes with name {list(skipped_names)} were not found in the NNCFGraph. " + error_msg
            )
        nncf_logger.info(f"{len(matched_by_names)} ignored nodes were found by name in the NNCFGraph")

    matched_by_patterns = []
    if ignored_scope.patterns:
        not_matched_patterns = []
        for str_pattern in ignored_scope.patterns:
            pattern = re.compile(str_pattern)
            matches = list(filter(pattern.match, node_names))
            if not matches:
                not_matched_patterns.append(str_pattern)
            matched_by_patterns.extend(matches)
        if strict and not_matched_patterns:
            raise nncf.ValidationError(
                f"No matches for ignored patterns {not_matched_patterns} in the NNCFGraph. " + error_msg
            )
        nncf_logger.info(f"{len(matched_by_patterns)} ignored nodes were found by patterns in the NNCFGraph")

    matched_by_types = []
    if ignored_scope.types:
        types_found = set()
        for node in nncf_graph.get_all_nodes():
            if node.node_type in ignored_scope.types:
                types_found.add(node.node_type)
                matched_by_types.append(node.node_name)
        not_matched_types = set(ignored_scope.types) - types_found
        if strict and not_matched_types:
            raise nncf.ValidationError(
                f"Nodes with ignored types {list(not_matched_types)} were not found in the NNCFGraph. " + error_msg
            )
        nncf_logger.info(f"{len(matched_by_types)} ignored nodes were found by types in the NNCFGraph")

    matched_by_subgraphs = []
    if ignored_scope.subgraphs:
        for subgraph in ignored_scope.subgraphs:
            names_from_subgraph = get_ignored_node_names_from_subgraph(nncf_graph, subgraph)
            if strict and not names_from_subgraph:
                raise nncf.ValidationError(
                    f"Ignored subgraph with input names {subgraph.inputs} and output names {subgraph.outputs} "
                    "was not found in the NNCFGraph. " + error_msg
                )

            matched_by_subgraphs.extend(names_from_subgraph)

    return set(matched_by_names + matched_by_types + matched_by_patterns + matched_by_subgraphs)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, Iterable, List, Optional, Tuple

import numpy as np
import openvino as ov
from openvino.runtime import opset13 as opset

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.utils import get_reduction_axes
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.experimental.tensor.tensor import Tensor
from nncf.openvino.graph.metatypes import openvino_metatypes as om
from nncf.openvino.graph.model_transformer import OVModelTransformer
from nncf.openvino.graph.node_utils import get_const_value
from nncf.openvino.graph.node_utils import get_weight_channel_axes
from nncf.openvino.graph.transformations.commands import OVTargetPoint
from nncf.openvino.rt_info import dump_parameters
from nncf.openvino.statistics.collectors import get_raw_stat_collector
from nncf.parameters import CompressWeightsMode
from nncf.quantization.algorithms.weight_compression.awq_patterns import get_awq_patterns
from nncf.quantization.algorithms.weight_compression.backend import WeightCompressionAlgoBackend
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionParameters
from nncf.quantization.algorithms.weight_compression.weight_lowering import compress_weight


class OVWeightCompressionAlgoBackend(WeightCompressionAlgoBackend):
    def __init__(self, model: ov.Model, name_to_node_mapping: Dict = None):
        if name_to_node_mapping is None:
            self.name_to_node_mapping = OVModelTransformer._get_name_to_node_mapping(model)
        else:
            self.name_to_node_mapping = name_to_node_mapping

    @property
    def matmul_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVMatMulMetatype]

    @property
    def convolution_metatypes(self) -> List[OperatorMetatype]:
        return [
            om.OVConvolutionMetatype,
            om.OVDepthwiseConvolutionMetatype,
            om.OVConvolutionBackpropDataMetatype,
            om.OVGroupConvolutionMetatype,
            om.OVGroupConvolutionBackpropDataMetatype,
        ]

    @property
    def embedding_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVEmbeddingMetatype]

    @staticmethod
    def is_node_with_weights(node: NNCFNode, graph: NNCFGraph) -> bool:
        return node.layer_attributes and node.layer_attributes.constant_attributes

    @staticmethod
    def get_reduction_axes(node_with_weight: NNCFNode, weight_port_id: int, graph: NNCFGraph) -> Optional[Tuple[int]]:
        channel_axes = get_weight_channel_axes(node_with_weight)
        const_shape = node_with_weight.layer_attributes.constant_attributes[weight_port_id]["shape"]
        return get_reduction_axes(channel_axes, const_shape)

    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> OVTargetPoint:
        return OVTargetPoint(target_type, target_node_name, port_id)

    @staticmethod
    def raw_statistic_collector(num_samples: Optional[int] = None) -> TensorCollector:
        return get_raw_stat_collector(num_samples)

    @staticmethod
    def get_activation_port_id(node: NNCFNode, nncf_graph: NNCFGraph) -> int:
        constant_ports = node.layer_attributes.get_const_port_ids()
        activation_ports = [
            e.input_port_id for e in nncf_graph.get_input_edges(node) if e.input_port_id not in constant_ports
        ]
        assert len(activation_ports) == 1
        return activation_ports[0]

    @staticmethod
    def get_weight_names_and_port_ids(node: NNCFNode, graph: NNCFGraph) -> List[Tuple[str, int]]:
        result = []
        for weight_port_id in node.layer_attributes.get_const_port_ids():
            weight_name = node.layer_attributes.constant_attributes[weight_port_id]["name"]
            result.append((weight_name, weight_port_id))
        return result

    def get_weight(self, node_with_weight: NNCFNode, weight_port_id: int, model: ov.Model, graph: NNCFGraph) -> Tensor:
        weight_name = node_with_weight.layer_attributes.constant_attributes[weight_port_id]["name"]
        weight_node = self.name_to_node_mapping[weight_name]
        weight_tensor = get_const_value(weight_node)
        return Tensor(weight_tensor)

    def set_weight(
        self, node_with_weight: NNCFNode, weight_port_id: int, model: ov.Model, graph: NNCFGraph, weight: Tensor
    ):
        node_with_const = self.name_to_node_mapping[node_with_weight.node_name]

        const_port = node_with_const.input(weight_port_id)
        const_node = node_with_const.input_value(weight_port_id).get_node()

        new_const_node = ov.runtime.op.Constant(weight.data, shared_memory=True)
        new_const_node.set_friendly_name(const_node.get_friendly_name())
        const_port.replace_source_output(new_const_node.output(0))

        const_name = node_with_weight.layer_attributes.constant_attributes[weight_port_id]["name"]
        self.name_to_node_mapping[const_name] = new_const_node

        new_output = new_const_node.output(0)
        for target_input in const_node.output(0).get_target_inputs():
            target_input.replace_source_output(new_output)

        del const_node

    def transform_model(
        self,
        model: ov.Model,
        graph: NNCFGraph,
        weight_compression_parameters: Iterable[WeightCompressionParameters],
        precomputed_scales: Dict[str, Tensor] = None,
        precomputed_zero_points: Dict[str, Tensor] = None,
    ) -> ov.Model:
        for wc_params in weight_compression_parameters:
            compression_config = wc_params.compression_config
            if compression_config.mode == CompressWeightsMode.NF4:
                compression_dtype = ov.Type.nf4
            elif compression_config.mode in [
                CompressWeightsMode.INT8_ASYM,
                CompressWeightsMode.INT8_SYM,
                CompressWeightsMode.INT8,
                CompressWeightsMode.INT4_ASYM,
                CompressWeightsMode.INT4_SYM,
            ]:
                if compression_config.mode in [CompressWeightsMode.INT4_ASYM, CompressWeightsMode.INT4_SYM]:
                    compression_dtype = ov.Type.u4
                else:
                    compression_dtype = ov.Type.u8
            else:
                raise ValueError(f"{compression_config.mode.value} is not supported.")

            const_attributes = wc_params.node_with_weight.layer_attributes.constant_attributes[wc_params.weight_port_id]
            const_node_name = const_attributes["name"]
            const_node = self.name_to_node_mapping[const_node_name]
            const_node_output = const_node.output(0)
            const_dtype = const_node_output.get_element_type()

            should_add_convert_node = False
            if const_dtype != ov.Type.f16:
                for inp in const_node_output.get_target_inputs():
                    if inp.get_node().get_type_name() != "Convert":
                        should_add_convert_node = True
                        break

            weight = Tensor(get_const_value(const_node, np.float32 if const_dtype == ov.Type.bf16 else None))
            original_shape = weight.shape
            compressed_weight = compress_weight(
                weight,
                wc_params.reduction_axes,
                compression_config,
                None if precomputed_scales is None else precomputed_scales.get(wc_params.weight_name),
                None if precomputed_zero_points is None else precomputed_zero_points.get(wc_params.weight_name),
            )

            compressed_const = opset.constant(
                compressed_weight.tensor.data, dtype=compression_dtype, name=const_node_name
            )
            converted_const = opset.convert(compressed_const, ov.Type.f16)
            if compressed_weight.zero_point is not None:
                zero_point_const = opset.constant(
                    compressed_weight.zero_point.data,
                    dtype=compression_dtype,
                    name=f"{const_node_name}/zero_point",
                )
                converted_zero_point = opset.convert(zero_point_const, ov.Type.f16)
                converted_const = opset.subtract(
                    converted_const, converted_zero_point, name=f"{const_node_name}/zero_point/subtract"
                )

            scale_const = opset.constant(
                compressed_weight.scale.data, dtype=ov.Type.f16, name=f"{const_node_name}/scale"
            )
            mul = opset.multiply(
                converted_const,
                scale_const,
                name=f"{const_node_name}/fq_weights_{wc_params.weight_port_id}",
            )

            if compression_config.group_size != -1:
                mul = opset.reshape(mul, output_shape=original_shape, special_zero=False)

            if should_add_convert_node:
                mul = opset.convert(
                    mul, const_dtype, name=f"{const_node_name}/fq_weights_{wc_params.weight_port_id}/convert"
                )

            mul_output = mul.output(0)
            for target_input in const_node.output(0).get_target_inputs():
                target_input.replace_source_output(mul_output)

        # reset name_to_node_mapping
        self.name_to_node_mapping = None

        return model

    @staticmethod
    def dump_parameters(
        model: ov.Model, parameters: Dict, algo_name: Optional[str] = "quantization", path: Optional[List] = None
    ) -> None:
        dump_parameters(model, parameters, algo_name, path)

    @staticmethod
    def get_compress_decompress_pipeline(
        weight_compression_parameter: WeightCompressionParameters, w_shape, s_shape, z_p_shape
    ):
        (
            w,
            s,
            zp,
            clamp,
        ) = OVWeightCompressionAlgoBackend.get_compress_pipeline(
            weight_compression_parameter, w_shape, s_shape, z_p_shape, True
        )

        result = (clamp - zp) * s
        model = ov.Model([result], [w, s, zp])

        compiled_model = ov.compile_model(model)

        return lambda w, s, zp: compiled_model([w, s, zp])[0]

    @staticmethod
    def get_compress_pipeline(
        weight_compression_parameter: WeightCompressionParameters, w_shape, s_shape, z_p_shape, return_nodes=False
    ):
        config = weight_compression_parameter.compression_config
        mode = config.mode
        assert mode in [CompressWeightsMode.INT4_SYM, CompressWeightsMode.INT4_ASYM]
        num_bits = config.num_bits

        level_low = 0
        level_high = 2**num_bits - 1

        w = opset.parameter(w_shape, name="w")
        s = opset.parameter(s_shape, name="s")
        zp = opset.parameter(z_p_shape, name="zp")

        result = opset.clamp(opset.round(w / s + zp), level_low, level_high, name="compressed_weights")

        if return_nodes:
            return w, s, zp, result

        model = ov.Model([result], [w, s, zp])

        compiled_model = ov.compile_model(model)

        return lambda w, s, zp: compiled_model([w, s, zp])[0]


class OVAWQAlgoAlgoBackend(OVWeightCompressionAlgoBackend):
    @staticmethod
    def get_awq_patterns():
        return get_awq_patterns(om.OVMatMulMetatype, om.OVMultiplyMetatype)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from typing import Callable, Dict, List, Optional, Set, Tuple, Type

import torch
from torch import nn

import nncf
from nncf.common.logging import nncf_logger
from nncf.common.scopes import matches_any
from nncf.torch.dynamic_graph.scope import Scope
from nncf.torch.dynamic_graph.scope import ScopeElement
from nncf.torch.layers import NNCF_MODULES
from nncf.torch.layers import NNCF_MODULES_DICT
from nncf.torch.layers import NNCF_MODULES_MAP
from nncf.torch.layers import NNCF_WRAPPED_USER_MODULES_DICT
from nncf.torch.layers import UNWRAPPED_USER_MODULES
from nncf.torch.layers import add_nncf_functionality_to_user_module
from nncf.torch.utils import get_model_device


def is_nncf_module(module: nn.Module) -> bool:
    """
    Checks whether given module has been extended with NNCF-enabling functionality.

    :param module: Module to check.
    :returns: True if module is an NNCF-extended module.
    """
    for nncf_module_name in NNCF_MODULES:
        if module.__class__.__name__ == nncf_module_name:
            return True
    for nncf_user_wrapped_class in NNCF_WRAPPED_USER_MODULES_DICT.values():
        if module.__class__.__name__ == nncf_user_wrapped_class.__name__:
            return True

    return False


def collect_all_scopes_for_extendable_and_extended_modules(
    model: nn.Module, predicate: Callable = None
) -> Dict[nn.Module, Set[Scope]]:
    """
    Collects all ranges for all modules in the model that match the condition from predicate.

    :param module: The model.
    :param predicate: A predicate function that can be used to filter modules.
    By default, the predicate function filters all NNCF modules and modules that can be replaced with NNCF modules.
    :return: A dictionary mapping modules to sets of scopes.
    """
    retval = {}
    if predicate is None:
        predicate = lambda x: _can_extend(x) or is_nncf_module(x)
    return _collect_modules_and_scopes_recursive_helper(model, Scope(), predicate, retval)


def collect_modules_and_scopes_by_predicate(
    module: nn.Module, predicate: Callable[[torch.nn.Module], bool]
) -> Dict[nn.Module, Set[Scope]]:
    retval = {}
    return _collect_modules_and_scopes_recursive_helper(module, Scope(), predicate, retval)


def _collect_modules_and_scopes_recursive_helper(
    current_module: nn.Module,
    current_scope: Scope,
    collect_predicate: Callable[[torch.nn.Module], bool],
    retval: Dict[nn.Module, Set[Scope]],
    visited_scopes: Set[Scope] = None,
) -> Dict[nn.Module, Set[Scope]]:
    if visited_scopes is None:
        visited_scopes = set()
        current_scope = Scope()
        current_scope.push(ScopeElement(current_module.__class__.__name__))

    if current_scope in visited_scopes:
        return retval

    visited_scopes.add(current_scope)

    for name, child_module in current_module.named_children():
        child_scope_element = ScopeElement(child_module.__class__.__name__, name)
        child_scope = current_scope.copy()
        child_scope.push(child_scope_element)

        if collect_predicate(child_module):
            if child_module not in retval:
                retval[child_module] = {child_scope}
            else:
                retval[child_module].add(child_scope)
        _ = _collect_modules_and_scopes_recursive_helper(
            child_module, child_scope, collect_predicate, retval, visited_scopes
        )

    return retval


def _can_extend(module: nn.Module) -> bool:
    """
    Whether the module can be replaced by an NNCF-extended version to enable compression via NNCF.
    :param module: Candidate module for replacement
    :return: Whether the module should be replaced.
    """
    return (
        module.__class__ in NNCF_MODULES_DICT.values()
        or module.__class__ in UNWRAPPED_USER_MODULES.registry_dict.values()
    )


def nncf_module_from(module: nn.Module) -> nn.Module:
    """
    Returns an NNCF-extended module from a given module that NNCF knows how to extend.

    :param module: The module to be replaced. Must be registered in NNCF as replaceable.
    :returns: The module extended with NNCF functionality.
    """
    assert _can_extend(module)
    for nncf_module_class, original_module_class in NNCF_MODULES_DICT.items():
        if module.__class__ == original_module_class:
            return nncf_module_class.from_module(module)
    for user_module_class in UNWRAPPED_USER_MODULES.registry_dict.values():
        if module.__class__ == user_module_class:
            nncf_module = deepcopy(module)
            nncf_module = add_nncf_functionality_to_user_module(nncf_module)
            return nncf_module
    raise nncf.InternalError(f"Could not extend module {module} with NNCF functionality!")


def replace_modules_by_nncf_modules(
    model: nn.Module,
    ignored_scopes: Optional[List[str]] = None,
    target_scopes: Optional[List[str]] = None,
    eval_op_scopes: Optional[List[Scope]] = None,
    custom_replacer: Callable[[nn.Module], None] = None,
    predicate_fn: Optional[Callable] = None,
) -> Tuple[nn.Module, Dict[torch.nn.Module, List[Scope]]]:
    """
    Replaces certain modules in the model hierarchy with NNCF-wrapped versions of the same modules.
    The modules to be replaced are statically defined in `nncf.torch.layers.NNCF_MODULES_DICT` and dynamically
    extended if the user utilized the `nncf.torch.layers.register_module` decorator to enable their custom
    weighted modules.
    Does not replace sub-modules of an already replaced module. Modules targeted via "target_scopes" will not
    be replaced if they are not already in another module targeted by "target_scopes".
    Note: for `ignored_scopes`, `target_scopes` and `eval_op_scopes`, the scopes
    in this list must correspond to the *storage* scope, i.e. reflect the hierarchy of the modules in the Python
    object,
    rather than correspond to the way that the underlying forward operations were executed in, which is different
    from the way other `ignored_scopes` and `target_scopes` are written in NNCF.

    :param model: The model in which the modules should be replaced.
    :param ignored_scopes: The list of string representations of the scopes that should be ignored during replacement,
     i.e. in which the replacement should not occur (corresponds to a "denylist").
    :param target_scopes: The list of string representations of the scopes so that the replacement should occur only
     in these scopes (corresponds to an "allowlist")
    :param eval_op_scopes: The list of the scopes for the modules that are executed in evaluation mode (i.e. the modules
    that end up having a scope not in this list will be considered train-only and will not be replaced).
    :param custom_replacer: The function to be used instead of the regular approach to replace a module with NNCF-
      extended counterpart.
    :param predicate_fn: The function to find modules that can be replaced.
    :return: The model with the modules replaced and the dictionary of all extended modules vs list of scopes through
    which the module is accessible. The list of scope shall be sorted lexicographically w.r.t. the string representation
    of the Scope objects.
    The dictionary will also include the extended modules that have already been present in the model.
    """
    modules_vs_scopes_dict = collect_all_scopes_for_extendable_and_extended_modules(model, predicate=predicate_fn)
    inter_dict: Dict[nn.Module, Set[Scope]] = {}
    ret_dict: Dict[nn.Module, List[Scope]] = {}
    for module, scope_set in modules_vs_scopes_dict.items():
        if is_nncf_module(module):
            # The module has already been extended, track it in the return value
            ret_dict[module] = list(sorted(scope_set, key=str))
            continue
        should_process = _is_scopes_allow_replacement(
            scope_set, ignored_scopes, target_scopes, eval_op_scopes
        ) and not _is_module_only_in_user_module(scope_set)
        if should_process:
            device = get_model_device(module)

            if custom_replacer is not None:
                replaced_module = custom_replacer(module)
            else:
                replaced_module = nncf_module_from(module)

            replaced_module.to(device)
            inter_dict[replaced_module] = scope_set

            new_scope_set = set()
            for scope in scope_set:
                # Adjust the returned scope so that it has the "NNCF..." version of the module
                new_scope = scope.copy()
                new_scope[-1].calling_module_class_name = replaced_module.__class__.__name__
                new_scope_set.add(new_scope)
                ret_dict[replaced_module] = list(sorted(new_scope_set, key=str))

    for replaced_module, old_scope_set in inter_dict.items():
        for old_scope in old_scope_set:
            # Should replace by all available scopes, otherwise some sub-containers may keep references to old
            # modules
            _replace_module_by_scope(model, old_scope, replaced_module)
    return model, ret_dict


def get_original_module_scope_from_nncf_module_scope(nncf_module_scope: Scope) -> Scope:
    original_module_scope = nncf_module_scope.copy()
    elt = original_module_scope.pop()
    assert elt.calling_module_class_name in NNCF_MODULES_MAP
    elt.calling_module_class_name = NNCF_MODULES_MAP[elt.calling_module_class_name]
    original_module_scope.push(elt)
    return original_module_scope


def _is_module_only_in_user_module(scope_set: Set[Scope]) -> bool:
    for scope in scope_set:
        has_at_least_one_user_module = False
        for user_module_class in UNWRAPPED_USER_MODULES.registry_dict.values():
            if _has_user_module(scope, user_module_class) or _has_user_module(
                scope, NNCF_WRAPPED_USER_MODULES_DICT[user_module_class]
            ):
                has_at_least_one_user_module = True
        if not has_at_least_one_user_module:
            # At least one scope from the input scope_set is located outside user module
            return False
    return True


def _has_user_module(scope: Scope, user_module_class: Type[torch.nn.Module]) -> bool:
    for scope_element in scope.scope_elements[:-1]:
        if scope_element.calling_module_class_name == user_module_class.__name__:
            return True
    return False


def _replace_module_by_scope(base_model: torch.nn.Module, scope: Scope, replaced_module: torch.nn.Module):
    """
    Accesses the module pointed to by scope in the base model and replaces it with an NNCF-extended module.

    :param base_model: The model relative to which the scope is built.
    :param scope: Scope of the module in base_module to be extended with NNCF functionality.
    :param custom_extender: The module that will replace the one pointed to by scope.
    :param reset: Whether the module should be reset.
    """
    curr_module = base_model
    owning_module = base_model
    for scope_element in scope[1:]:  # omit first scope element which corresponds to base module
        child_module = curr_module._modules.get(scope_element.calling_field_name)
        if child_module is None:
            raise nncf.InternalError(
                "Could not find a {} module member in {} module of scope {} during module replacement".format(
                    scope_element.calling_field_name, scope_element.calling_module_class_name, str(scope)
                )
            )
        owning_module = curr_module
        curr_module = child_module

    if is_nncf_module(curr_module):
        # Already replaced, possibly via another scope alias
        nncf_logger.debug(f"Module at {str(scope)} was already extended.")
        return

    nncf_logger.debug(f"Extending module at {str(scope)}...")
    last_calling_field_name = scope[-1].calling_field_name
    if isinstance(owning_module, nn.Sequential):
        owning_module._modules[last_calling_field_name] = replaced_module
    else:
        setattr(owning_module, last_calling_field_name, replaced_module)


def _is_scopes_allow_replacement(
    scope_set_for_module: Set[Scope],
    ignored_scopes: Optional[List[str]] = None,
    target_scopes: Optional[List[str]] = None,
    eval_op_scopes: Optional[List[Scope]] = None,
) -> bool:
    used_in_eval = False
    for scope in scope_set_for_module:
        if matches_any(str(scope), ignored_scopes):
            nncf_logger.info(
                f"Not processing a module that matched to an ignored scope in config; module scope = {str(scope)}"
            )
            return False
        if eval_op_scopes is not None:
            for eval_op_scope in eval_op_scopes:
                # child_scope isn't ignored, if there's at least a single operation or a module called in eval mode
                # inside it
                if eval_op_scope in scope:
                    used_in_eval = True
                    break

    if eval_op_scopes is not None and not used_in_eval:
        return False

    if target_scopes is not None:
        is_any_scope_in_target = False
        for scope in scope_set_for_module:
            if matches_any(str(scope), target_scopes):
                is_any_scope_in_target = True
                break
        if not is_any_scope_in_target:
            nncf_logger.info(
                f"Not processing a module outside target scope specified in config; "
                f"module known under scope(s) = {';'.join([str(x) for x in scope_set_for_module])}"
            )
            return False

    return True



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import singledispatch
from typing import Dict

from nncf.common.pruning.statistics import FilterPruningStatistics
from nncf.common.sparsity.statistics import ConstSparsityStatistics
from nncf.common.sparsity.statistics import MagnitudeSparsityStatistics
from nncf.common.sparsity.statistics import MovementSparsityStatistics
from nncf.common.sparsity.statistics import RBSparsityStatistics
from nncf.common.statistics import NNCFStatistics


def prepare_for_tensorboard(nncf_stats: NNCFStatistics) -> Dict[str, float]:
    """
    Extracts scalar values from NNCF statistics for its reporting to the TensorBoard.

    :param nncf_stats: NNCF Statistics.
    :return: A dict storing name and value of the scalar.
    """
    tensorboard_stats = {}
    for algorithm_name, stats in nncf_stats:
        tensorboard_stats.update(convert_to_dict(stats, algorithm_name))

    return tensorboard_stats


@singledispatch
def convert_to_dict(stats, algorithm_name: str):
    return {}


@convert_to_dict.register(FilterPruningStatistics)
def _(stats, algorithm_name):
    tensorboard_stats = {
        f"{algorithm_name}/algo_current_pruning_level": stats.current_pruning_level,
        f"{algorithm_name}/model_FLOPS_pruning_level": stats.model_statistics.flops_pruning_level,
        f"{algorithm_name}/model_params_pruning_level": stats.model_statistics.params_pruning_level,
        f"{algorithm_name}/model_filters_pruning_level": stats.model_statistics.filter_pruning_level,
    }
    return tensorboard_stats


@convert_to_dict.register(MagnitudeSparsityStatistics)
@convert_to_dict.register(RBSparsityStatistics)
@convert_to_dict.register(ConstSparsityStatistics)
def _(stats, algorithm_name):
    tensorboard_stats = {
        f"{algorithm_name}/sparsity_level_for_model": stats.model_statistics.sparsity_level,
        f"{algorithm_name}/sparsity_level_for_sparsified_layers": stats.model_statistics.sparsity_level_for_layers,
    }

    target_sparsity_level = getattr(stats, "target_sparsity_level", None)
    if target_sparsity_level is not None:
        tensorboard_stats[f"{algorithm_name}/target_sparsity_level"] = target_sparsity_level

    return tensorboard_stats


@convert_to_dict.register(MovementSparsityStatistics)
def _(stats, algorithm_name):
    tensorboard_stats = {
        f"{algorithm_name}/model_sparsity": stats.model_statistics.sparsity_level,
        f"{algorithm_name}/linear_layer_sparsity": stats.model_statistics.sparsity_level_for_layers,
        f"{algorithm_name}/importance_threshold": stats.importance_threshold,
        f"{algorithm_name}/importance_regularization_factor": stats.importance_regularization_factor,
    }
    return tensorboard_stats



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
import json
from bisect import bisect_left
from collections import OrderedDict
from copy import deepcopy
from enum import Enum
from operator import itemgetter
from pathlib import Path
from typing import Any, Callable, Dict, List, NamedTuple, Set, Tuple

import torch
from torch import Tensor
from torch import nn
from torch.nn.modules.loss import _Loss

import nncf
from nncf.common.graph import NNCFNodeName
from nncf.common.logging import nncf_logger
from nncf.common.quantization.quantizer_setup import QuantizationPointId
from nncf.common.quantization.quantizer_setup import SingleConfigQuantizerSetup
from nncf.common.quantization.structs import QuantizerConfig
from nncf.common.quantization.structs import QuantizerId
from nncf.common.quantization.structs import WeightQuantizerId
from nncf.common.utils.debug import is_debug
from nncf.common.utils.os import safe_open
from nncf.config.schemata.defaults import HAWQ_COMPRESSION_RATIO
from nncf.config.schemata.defaults import HAWQ_DUMP_INIT_PRECISION_DATA
from nncf.config.schemata.defaults import HAWQ_ITER_NUMBER
from nncf.config.schemata.defaults import HAWQ_NUM_DATA_POINTS
from nncf.config.schemata.defaults import HAWQ_TOLERANCE
from nncf.config.schemata.defaults import PRECISION_INIT_BITWIDTHS
from nncf.torch.quantization.hessian_trace import HessianTraceEstimator
from nncf.torch.quantization.layers import QuantizersSwitcher
from nncf.torch.quantization.precision_constraints import HardwareQuantizationConstraints
from nncf.torch.quantization.precision_init.adjacent_quantizers import GroupsOfAdjacentQuantizers
from nncf.torch.quantization.precision_init.base_init import BasePrecisionInitializer
from nncf.torch.quantization.precision_init.base_init import BasePrecisionInitParams
from nncf.torch.quantization.precision_init.compression_ratio import CompressionRatioCalculator
from nncf.torch.quantization.precision_init.definitions import CoveringQConfigSequenceForQuantNoiseCalculation
from nncf.torch.quantization.precision_init.definitions import QConfigSequenceForHAWQToEvaluate
from nncf.torch.quantization.precision_init.hawq_debug import HAWQDebugger
from nncf.torch.quantization.precision_init.perturbations import PerturbationObserver
from nncf.torch.quantization.precision_init.perturbations import Perturbations
from nncf.torch.quantization.precision_init.traces_order import TracesOrder
from nncf.torch.quantization.precision_init.traces_order import TracesPerLayer
from nncf.torch.quantization.structs import WeightQuantizerInfo
from nncf.torch.structures import QuantizationPrecisionInitArgs
from nncf.torch.utils import get_model_device


class BitwidthAssignmentMode(Enum):
    STRICT = "strict"
    LIBERAL = "liberal"


class HAWQPrecisionInitParams(BasePrecisionInitParams):
    def __init__(
        self,
        user_init_args: QuantizationPrecisionInitArgs,
        bitwidths: List[int] = None,
        bitwidth_per_scope: List[List] = None,
        traces_per_layer_path: str = None,
        num_data_points: int = None,
        iter_number: int = None,
        tolerance: float = None,
        compression_ratio: float = None,
        dump_hawq_data: bool = None,
        bitwidth_assignment_mode: BitwidthAssignmentMode = None,
    ):
        super().__init__(user_init_args)
        self.bitwidths = bitwidths
        self.bitwidth_per_scope = bitwidth_per_scope
        self.traces_per_layer_path = traces_per_layer_path
        self.num_data_points = num_data_points
        self.iter_number = iter_number
        self.tolerance = tolerance
        self.compression_ratio = compression_ratio
        self.dump_hawq_data = dump_hawq_data
        self.bitwidth_assignment_mode = bitwidth_assignment_mode

    @classmethod
    def from_config(
        cls, hawq_init_config_dict: Dict, user_init_args: QuantizationPrecisionInitArgs
    ) -> "HAWQPrecisionInitParams":
        return cls(
            user_init_args=user_init_args,
            bitwidths=hawq_init_config_dict.get("bits", PRECISION_INIT_BITWIDTHS),
            traces_per_layer_path=hawq_init_config_dict.get("traces_per_layer_path"),
            num_data_points=hawq_init_config_dict.get("num_data_points", HAWQ_NUM_DATA_POINTS),
            iter_number=hawq_init_config_dict.get("iter_number", HAWQ_ITER_NUMBER),
            tolerance=hawq_init_config_dict.get("tolerance", HAWQ_TOLERANCE),
            compression_ratio=hawq_init_config_dict.get("compression_ratio", HAWQ_COMPRESSION_RATIO),
            dump_hawq_data=hawq_init_config_dict.get("dump_init_precision_data", HAWQ_DUMP_INIT_PRECISION_DATA),
            bitwidth_assignment_mode=BitwidthAssignmentMode(
                hawq_init_config_dict.get("bitwidth_assignment_mode", BitwidthAssignmentMode.LIBERAL.value)
            ),
        )


class TraceOrderBitwidthMatcher:
    def __init__(self, available_bitwidths: List[int], traces_order: TracesOrder):
        self._available_bitwidths = available_bitwidths
        self._traces_order = traces_order
        self._bitwidth_sequences = self.get_all_non_decreasing_bitwidth_sequences()

    def get_all_non_decreasing_bitwidth_sequences(self) -> List[List[int]]:
        sequences = []
        bitwidths_ = deepcopy(self._available_bitwidths)
        seq_len = len(self._traces_order)
        if seq_len == 0:
            return sequences
        bitwidths = sorted(bitwidths_)
        m = len(bitwidths)
        L = seq_len
        for j in range(1, m + 1):
            for combo_bitwidths in itertools.combinations(bitwidths, j):
                for combo_partitions in itertools.combinations(list(range(1, L)), j - 1):
                    bit_config = []
                    prev_p = 0
                    for p, b in zip(combo_partitions + (L,), combo_bitwidths):
                        bit_config += [b] * (p - prev_p)
                        prev_p = p
                    sequences.append(bit_config)
        return sequences

    @staticmethod
    def _select_first_closest_bitwidth_qconfig(
        qconf_list: List[QuantizerConfig], target_bitwidth: int
    ) -> QuantizerConfig:
        bw_diffs = [abs(qc.num_bits - target_bitwidth) for qc in qconf_list]
        _, min_idx = min((val, idx) for (idx, val) in enumerate(bw_diffs))
        return qconf_list[min_idx]

    def _deduplicate(
        self, qconf_sequences_to_search: List[QConfigSequenceForHAWQToEvaluate]
    ) -> List[QConfigSequenceForHAWQToEvaluate]:
        tupled_sequence = [tuple(seq) for seq in qconf_sequences_to_search]
        odict = OrderedDict.fromkeys(tupled_sequence)
        deduped_tupled_sequence = list(odict.keys())
        return [list(tup) for tup in deduped_tupled_sequence]

    @staticmethod
    def _generate_covering_qconfig_sequences(observed_qconfs: List[Dict[QuantizerConfig, QuantizerConfig]]):
        covering_qconfig_sequences: List[CoveringQConfigSequenceForQuantNoiseCalculation] = []
        # For each index, put the largest qconf subset that only varies in bitwidth on top
        # so that the associated covering configurations would not require model regeneration
        optimized_observed_qconfs: List[List[QuantizerConfig]] = []
        for qconf_oset in observed_qconfs:
            variants: List[List[QuantizerConfig]] = []
            for qconf in qconf_oset:
                variants.append(list(filter(qconf.is_a_bitwidth_variant, qconf_oset.keys())))
            max_bw_varying_variant = max(variants, key=len)
            other_qconfs = list(filter(lambda x: x not in max_bw_varying_variant, qconf_oset.keys()))
            optimized_observed_qconfs.append(max_bw_varying_variant + other_qconfs)

        max_depth = max([len(qconfs_for_trace_idx) for qconfs_for_trace_idx in optimized_observed_qconfs])
        for i in range(max_depth):
            covering_conf: CoveringQConfigSequenceForQuantNoiseCalculation = []
            for qconfs_for_trace_idx in optimized_observed_qconfs:
                if i < len(qconfs_for_trace_idx):
                    covering_conf.append(qconfs_for_trace_idx[i])
                else:
                    covering_conf.append(qconfs_for_trace_idx[-1])
            covering_qconfig_sequences.append(covering_conf)
        return covering_qconfig_sequences

    def get_qconfig_sequences_constrained_by_trace_order(
        self,
        possible_qconfigs_sequence_in_trace_order: List[List[QuantizerConfig]],
        indices_for_bitwidth_adjustment_only: Set[int],
    ) -> Tuple[List[QConfigSequenceForHAWQToEvaluate], List[CoveringQConfigSequenceForQuantNoiseCalculation]]:
        """
        The 'constraint' is so that the each qconfig sequence should have non-decreasing bitwidths. It
        might be impossible to apply this constraint for a given qconfig space (consider [[2], [6, 8], [4]]).
        In such a case, for trace order index positions where it was impossible to select a bitwidth so that the entire
        sequence is non-decreasing, the bitwidth closest to this target will be chosen instead.
        """
        if len(possible_qconfigs_sequence_in_trace_order) != len(self._traces_order):
            raise ValueError("The size of the qconfig space and the traces do not match!")
        retval: List[QConfigSequenceForHAWQToEvaluate] = []
        observed_qconfs_in_retval = [OrderedDict() for _ in range(len(self._traces_order))]
        for bitwidth_sequence in self._bitwidth_sequences:
            current_qconfig_sequence_in_trace_order: QConfigSequenceForHAWQToEvaluate = []
            for trace_idx, bitwidth in enumerate(bitwidth_sequence):
                if trace_idx in indices_for_bitwidth_adjustment_only:
                    bitwidth_adjusted_default_qconfig = deepcopy(
                        possible_qconfigs_sequence_in_trace_order[trace_idx][0]
                    )
                    bitwidth_adjusted_default_qconfig.num_bits = bitwidth
                    qconfig = bitwidth_adjusted_default_qconfig
                else:
                    # TODO: do a selection based on strategy ("exhaustive" = add all available configurations,
                    # "preset" = do a selection based on a certain preset, "first" = select first match (as below),
                    # "custom" = use a custom selection function to be passed as arg to the HAWQ initializer
                    # OR: do non-bitwidth disambiguation higher up the stack, make sure that the qconfig
                    # space at this spot only has 1 qconfig option for each bitwidth.
                    possible_qconfigs_for_current_trace_idx = possible_qconfigs_sequence_in_trace_order[trace_idx]
                    first_closest_qconfig = self._select_first_closest_bitwidth_qconfig(
                        possible_qconfigs_for_current_trace_idx, bitwidth
                    )
                    qconfig = deepcopy(first_closest_qconfig)

                current_qconfig_sequence_in_trace_order.append(qconfig)
                observed_qconfs_in_retval[trace_idx][qconfig] = qconfig
            retval.append(current_qconfig_sequence_in_trace_order)
        return self._deduplicate(retval), self._generate_covering_qconfig_sequences(observed_qconfs_in_retval)


class HAWQPrecisionInitializer(BasePrecisionInitializer):
    def __init__(
        self,
        algo: "ExperimentalQuantizationController",  # noqa: F821
        params: HAWQPrecisionInitParams,
        hw_precision_constraints: HardwareQuantizationConstraints,
    ):
        self._groups_of_adjacent_quantizers = algo.groups_of_adjacent_quantizers
        self._bitwidth_assignment_mode = params.bitwidth_assignment_mode
        if self._bitwidth_assignment_mode == BitwidthAssignmentMode.STRICT:
            hw_precision_constraints = self._merge_constraints_for_adjacent_quantizers(
                self._groups_of_adjacent_quantizers, hw_precision_constraints
            )
        super().__init__(algo, params, hw_precision_constraints)
        init_args = params.user_init_args
        self._criterion_fn = init_args.criterion_fn
        self._criterion = init_args.criterion
        self._data_loader = init_args.data_loader
        self._traces_per_layer_path = params.traces_per_layer_path
        self._num_data_points = params.num_data_points
        self._iter_number = params.iter_number
        self._tolerance = params.tolerance
        self._compression_ratio = params.compression_ratio
        self._bitwidths = (
            self._hw_precision_constraints.get_all_unique_bitwidths()
            if self._hw_precision_constraints
            else params.bitwidths
        )
        self._init_device = init_args.device
        if self._init_device is None:
            self._init_device = get_model_device(self._model)
        current_quantizer_setup = self._algo.get_quantizer_setup_for_current_state()
        flops_per_module = self._model.nncf.get_flops_per_module()
        self._compression_ratio_calculator = CompressionRatioCalculator(
            flops_per_module,
            current_quantizer_setup,
            self._groups_of_adjacent_quantizers.weight_qp_id_per_activation_qp_id,
        )
        self._dump_hawq_data = params.dump_hawq_data
        self._original_qp_id_vs_quantizer_module_id_dict = deepcopy(algo.setup_to_module_id_translation_dict)

    def apply_init(self) -> SingleConfigQuantizerSetup:
        if not self._weight_quantizations_by_execution_order:
            return self._algo.get_quantizer_setup_for_current_state()

        original_device = get_model_device(self._model)
        self._model.to(self._init_device)

        traces_per_layer = self._calc_traces(self._criterion_fn, self._criterion, self._iter_number, self._tolerance)
        if not traces_per_layer:
            raise nncf.InternalError("Failed to calculate hessian traces!")

        traces_order = traces_per_layer.traces_order
        (
            weight_qconfig_sequences_in_trace_order,
            covering_qconfig_sequences,
        ) = self.get_qconfig_sequences_constrained_by_traces_order(traces_order)

        weight_quantizer_ids_in_execution_order = list(self._weight_quantizations_by_execution_order.keys())

        if not weight_qconfig_sequences_in_trace_order:
            nncf_logger.error("All bitwidths configurations are incompatible with HW Config!")
            return None

        weight_qconfig_sequences_in_trace_order = self._filter_qconfig_sequences_by_excessive_bitwidth(
            weight_qconfig_sequences_in_trace_order
        )

        if self._bitwidth_assignment_mode == BitwidthAssignmentMode.STRICT:
            weight_qconfig_sequences_in_trace_order = self._filter_qconfig_sequences_by_grouped_weight_quantizers(
                weight_qconfig_sequences_in_trace_order,
                weight_quantizer_ids_in_execution_order,
                self._groups_of_adjacent_quantizers,
                traces_order,
            )
        if not weight_qconfig_sequences_in_trace_order:
            nncf_logger.error(
                "No bitwidths configurations are left after removing inconsistent groups of "
                "weight quantizers with adjacent activation quantizers!"
            )
            return self._algo.get_quantizer_setup_for_current_state()

        compression_ratio_per_qconfig = self.get_compression_ratio_per_qconfig_sequence(
            weight_qconfig_sequences_in_trace_order, traces_order
        )
        min_ratio = min(compression_ratio_per_qconfig)
        max_ratio = max(compression_ratio_per_qconfig)
        if not min_ratio <= self._compression_ratio <= max_ratio:
            raise AttributeError(
                "Invalid compression ratio={}. Should be within range [{:.3f}, {:.3f}]".format(
                    self._compression_ratio, min_ratio, max_ratio
                )
            )

        perturbations, weight_observers = self.calc_quantization_noise(covering_qconfig_sequences, traces_order)

        metric_per_qconfig_sequence = self.calc_hawq_metric_per_qconfig_sequence(
            weight_qconfig_sequences_in_trace_order, perturbations, traces_per_layer, self._init_device
        )

        qconfig_sequence_index = self.choose_qconfig_sequence(
            metric_per_qconfig_sequence, compression_ratio_per_qconfig, self._compression_ratio
        )
        chosen_qconfig_sequence_in_traces_order = weight_qconfig_sequences_in_trace_order[qconfig_sequence_index]
        chosen_qconfig_sequence_in_execution_order = traces_order.get_execution_order_configs(
            chosen_qconfig_sequence_in_traces_order
        )
        bitwidth_sequence = [qconfig.num_bits for qconfig in chosen_qconfig_sequence_in_execution_order]
        nncf_logger.info(
            f"Chosen HAWQ bitwidth sequence with ratio={compression_ratio_per_qconfig[qconfig_sequence_index]:.2f}, "
            f"bitwidth per weightable layer={bitwidth_sequence}"
        )
        nncf_logger.debug(
            f"Order of the weightable layers in the HAWQ bitwidth sequence "
            f"(in descending order of average Hessian traces) = {traces_order}"
        )

        final_quantizer_setup = self.get_quantizer_setup_for_qconfig_sequence(
            chosen_qconfig_sequence_in_traces_order, traces_order
        )
        if is_debug() or self._dump_hawq_data:
            hawq_debugger = HAWQDebugger(
                weight_qconfig_sequences_in_trace_order,
                perturbations,
                weight_observers,
                traces_per_layer,
                self._bitwidths,
            )
            hawq_debugger.dump_metric_MB(metric_per_qconfig_sequence)
            hawq_debugger.dump_metric_flops(
                metric_per_qconfig_sequence, compression_ratio_per_qconfig, qconfig_sequence_index
            )
            hawq_debugger.dump_avg_traces()
            hawq_debugger.dump_density_of_quantization_noise()
            hawq_debugger.dump_perturbations_ratio()
            new_ctrl, new_model = self._algo.apply_new_quantizer_setup(final_quantizer_setup)
            groups_of_adjacent_quantizers = new_ctrl.groups_of_adjacent_quantizers
            hawq_debugger.dump_bitwidth_graph(new_ctrl, new_model, groups_of_adjacent_quantizers)
        bitwidth_per_scope = self.get_bitwidth_per_scope(final_quantizer_setup)
        from nncf.common.utils.debug import DEBUG_LOG_DIR

        Path(DEBUG_LOG_DIR).mkdir(parents=True, exist_ok=True)
        with safe_open(Path(DEBUG_LOG_DIR) / "bitwidth_per_scope.json", "w") as outfile:
            json.dump({"bitwidth_per_scope": bitwidth_per_scope}, outfile, indent=4, sort_keys=False)
        self._model.to(original_device)
        return final_quantizer_setup

    @staticmethod
    def _merge_constraints_for_adjacent_quantizers(
        groups_of_adjacent_quantizers: GroupsOfAdjacentQuantizers,
        hw_precision_constraints: HardwareQuantizationConstraints,
    ) -> HardwareQuantizationConstraints:
        if not hw_precision_constraints:
            return None
        retval = deepcopy(hw_precision_constraints)
        for group in groups_of_adjacent_quantizers:
            all_bitwidths_sets = []
            quantizer_ids = []
            all_quantizers = group.weight_quantizers + group.activation_quantizers
            for quantizer_id, _ in all_quantizers:
                bitwidths_vs_qconfig_sequence = retval.get_bitwidth_vs_qconfigs_dict(quantizer_id)
                bitwidths = set(bitwidths_vs_qconfig_sequence.keys())
                all_bitwidths_sets.append(bitwidths)
                quantizer_ids.append(quantizer_id)
            minimal_set_bitwidths = set.intersection(*all_bitwidths_sets)
            if not minimal_set_bitwidths:
                raise nncf.InternalError(
                    "No bitwidths configurations are left after removing inconsistent groups of weight quantizers"
                    " with adjacent activation quantizers!"
                )
            for quantizer_id in quantizer_ids:
                qconfig_sequence = retval.get(quantizer_id)
                filtered_qconfig_sequence = []
                for qconf in qconfig_sequence:
                    if qconf.num_bits in minimal_set_bitwidths:
                        filtered_qconfig_sequence.append(qconf)
                retval.replace(quantizer_id, filtered_qconfig_sequence)
        return retval

    def get_compression_ratio_per_qconfig_sequence(
        self, qconfig_sequences_in_trace_order: List[QConfigSequenceForHAWQToEvaluate], traces_order: TracesOrder
    ) -> List[float]:
        compression_ratio_per_qconfig = []
        for qconfig_sequence in qconfig_sequences_in_trace_order:
            quantizer_setup = self.get_quantizer_setup_for_qconfig_sequence(qconfig_sequence, traces_order)
            compression_ratio = self._compression_ratio_calculator.run_for_quantizer_setup(quantizer_setup)
            compression_ratio_per_qconfig.append(compression_ratio)
        return compression_ratio_per_qconfig

    class ParamsToRestore(NamedTuple):
        originally_disabled_gradients: List[str]
        skipped_gradients_to_enable: List[Tuple[nn.Module, str]]

    @staticmethod
    def disable_all_gradients_except_weights_of_quantized_modules(
        quantizers_switcher: QuantizersSwitcher,
        weight_quantizers: Dict[WeightQuantizerId, WeightQuantizerInfo],
        model: nn.Module,
        skipped_quantized_weight_node_names: List[NNCFNodeName] = None,
    ) -> ParamsToRestore:
        """
        Disables gradients of all parameters, except for layers that have quantizers for weights, which wasn't skipped
        because of single precision constraints.
        :param quantizers_switcher: object that is responsible for enabling and disabling quantizers
        :param weight_quantizers: modules with quantized weights per scope
        :param model: model to access all parameters
        :param skipped_quantized_weight_node_names: list of weighted nodes that have a single precision
        constraint and which weights should be skipped from bitwidth initialization
        :return: list of names of the parameters that were originally disabled
        """
        originally_disabled_gradients = []
        skipped_gradients_to_enable = []

        # Some quantizers can be disabled in a staged scenario on creation of staged scheduler
        # Need to save originally disabled quantizers for restoring their state after initialization
        quantizers_switcher.disable_quantizers()

        # remember gradients of quantized modules that were enabled
        gradients_to_enable = []
        for wq_id, wq_info in weight_quantizers.items():
            quantized_module = wq_info.quantized_module
            target_node_name = wq_id.target_node_name
            is_skipped = False
            for skipped_wt_node_name in skipped_quantized_weight_node_names:
                if skipped_wt_node_name == target_node_name:
                    is_skipped = True
                    break
            for param_name, param in quantized_module.named_parameters():
                if param.requires_grad:
                    # disable gradients for skipped module for optimization of Hessian Trace search
                    if is_skipped:
                        skipped_gradients_to_enable.append((quantized_module, param_name))
                        param.requires_grad = False
                    else:
                        gradients_to_enable.append((quantized_module, param_name))

        # disable all gradients, except already disabled
        for param_name, param in model.named_parameters():
            if not param.requires_grad:
                originally_disabled_gradients.append(param_name)
            else:
                param.requires_grad = False

        # enable gradients of quantized modules that were disabled
        for wq_id in weight_quantizers.values():
            quantized_module = wq_id.quantized_module
            for param_name, param in quantized_module.named_parameters():
                if (quantized_module, param_name) in gradients_to_enable and "bias" not in param_name:
                    param.requires_grad = True
        return HAWQPrecisionInitializer.ParamsToRestore(originally_disabled_gradients, skipped_gradients_to_enable)

    def _calc_traces(
        self,
        criterion_fn: Callable[[Any, Any, _Loss], torch.Tensor],
        criterion: _Loss,
        iter_number: int,
        tolerance: float,
    ) -> TracesPerLayer:
        if self._traces_per_layer_path:
            return TracesPerLayer(torch.load(self._traces_per_layer_path).to(self._init_device))

        quantizers_switcher = QuantizersSwitcher(list(self._all_quantizers_per_scope.values()))
        params_to_restore = self.disable_all_gradients_except_weights_of_quantized_modules(
            quantizers_switcher,
            self._algo.weight_quantizers,
            self._model,
            self._quantizers_handler.get_skipped_quantized_weight_node_names(),
        )

        trace_estimator = HessianTraceEstimator(
            self._model, criterion_fn, criterion, self._init_device, self._data_loader, self._num_data_points
        )
        try:
            avg_traces = trace_estimator.get_average_traces(max_iter=iter_number, tolerance=tolerance)
        except RuntimeError as error:
            if "cuda out of memory" in error.args[0].lower():
                raise nncf.InternalError(
                    "Failed to estimate average Hessian traces within precision initialization. Specify "
                    "a smaller batch size via --batch-size-init option in the NNCF samples or register "
                    "a data loader with a smaller batch size. Refer to "
                    "`NNCFConfig.register_extra_structs` and the `QuantizationPrecisionInitArgs`"
                    " class"
                ) from error
            raise error

        self.restore_disabled_gradients(
            quantizers_switcher, self._model, self._algo.weight_quantizers, params_to_restore
        )

        return TracesPerLayer(avg_traces)

    @staticmethod
    def restore_disabled_gradients(
        quantizers_switcher: QuantizersSwitcher,
        model: nn.Module,
        weight_quantizers: Dict[WeightQuantizerId, WeightQuantizerInfo],
        params_to_restore: ParamsToRestore,
    ):
        """
        Restore requires_grad property of all parameters back, except for ones that were originally disabled
        :param quantizers_switcher: object that is responsible for enabling and disabling quantizers
        :param model: model to access all parameters
        :param weight_quantizers: modules with quantized weights per scope
        :param params_to_restore: storage names of the parameters that should restore reguires_grad property
        """
        for wq_info in weight_quantizers.values():
            quantized_module = wq_info.quantized_module
            for param_name, param in quantized_module.named_parameters():
                if (quantized_module, param_name) in params_to_restore.skipped_gradients_to_enable:
                    param.requires_grad = True

        for param_name, param in model.named_parameters():
            if param_name not in params_to_restore.originally_disabled_gradients:
                param.requires_grad = True
        quantizers_switcher.enable_quantizers()

    def get_qconfig_sequences_constrained_by_traces_order(
        self, traces_order: TracesOrder
    ) -> Tuple[List[QConfigSequenceForHAWQToEvaluate], List[CoveringQConfigSequenceForQuantNoiseCalculation]]:
        possible_qconfigs_sequence_in_trace_order: List[List[QuantizerConfig]] = []
        trace_order_indices_of_defaulted_qconfig_sequence: Set[int] = set()
        quantizer_ids_in_exec_order = list(self._weight_quantizations_by_execution_order.keys())
        assert len(quantizer_ids_in_exec_order) == len(traces_order)
        for trace_idx in range(len(traces_order)):
            exec_idx = traces_order.get_execution_index_by_traces_index(trace_idx)
            qid = quantizer_ids_in_exec_order[exec_idx]
            default_qconfig = self._weight_quantizations_by_execution_order[qid].get_quantizer_config()
            qconfig_constraints = []
            if self._hw_precision_constraints:
                qconfig_constraints = self._hw_precision_constraints.get(qid)
            if qconfig_constraints:
                possible_qconfigs_sequence_in_trace_order.append(qconfig_constraints)
            else:
                possible_qconfigs_sequence_in_trace_order.append([default_qconfig])
                trace_order_indices_of_defaulted_qconfig_sequence.add(trace_idx)

        matcher = TraceOrderBitwidthMatcher(self._bitwidths, traces_order)
        return matcher.get_qconfig_sequences_constrained_by_trace_order(
            possible_qconfigs_sequence_in_trace_order, trace_order_indices_of_defaulted_qconfig_sequence
        )

    def _get_weight_qp_ids_in_trace_order(self, traces_order: TracesOrder) -> List[Set[QuantizationPointId]]:
        quant_module_ids = list(self._weight_quantizations_by_execution_order.keys())
        qp_ids_in_trace_order = []
        for trace_idx in range(len(traces_order)):
            exec_idx = traces_order.get_execution_index_by_traces_index(trace_idx)
            quant_module_id = quant_module_ids[exec_idx]
            qp_ids_in_trace_order.append(self._algo.module_id_to_qp_id_translation_dict[quant_module_id])
        return qp_ids_in_trace_order

    @staticmethod
    def _apply_qconfig_sequence_to_quantizer_setup(
        qconfig_sequence: CoveringQConfigSequenceForQuantNoiseCalculation,
        qp_ids_in_trace_order: List[Set[QuantizationPointId]],
        quantizer_setup: SingleConfigQuantizerSetup,
    ) -> SingleConfigQuantizerSetup:
        retval = deepcopy(quantizer_setup)
        assert len(qconfig_sequence) == len(qp_ids_in_trace_order)
        for trace_idx, qp_id_set in enumerate(qp_ids_in_trace_order):
            for qp_id in qp_id_set:
                retval.quantization_points[qp_id].qconfig = deepcopy(qconfig_sequence[trace_idx])
        return retval

    def calc_quantization_noise(
        self, qconfig_sequences_to_run: List[CoveringQConfigSequenceForQuantNoiseCalculation], traces_order: TracesOrder
    ) -> Tuple[Perturbations, List[List[PerturbationObserver]]]:
        perturbations = Perturbations()
        qp_ids_in_trace_order = self._get_weight_qp_ids_in_trace_order(traces_order)
        ctrl = self._algo
        observers_for_all_qconfig_sequences: List[List[PerturbationObserver]] = []
        for qconfig_sequence in qconfig_sequences_to_run:
            quantizer_setup_to_run = self._apply_qconfig_sequence_to_quantizer_setup(
                qconfig_sequence, qp_ids_in_trace_order, ctrl.get_quantizer_setup_for_current_state()
            )
            ctrl, model = ctrl.apply_new_quantizer_setup(quantizer_setup_to_run)

            hook_handles = []
            observers = []
            for qp_id_set in qp_ids_in_trace_order:
                for qp_id in qp_id_set:
                    wq_id = ctrl.setup_to_module_id_translation_dict[qp_id]
                    wq_module = ctrl.weight_quantizers[wq_id].quantizer_module_ref
                    observer = PerturbationObserver(self._init_device)
                    hook_handles.append(wq_module.register_forward_hook(observer.calc_perturbation))
                    observers.append(observer)

            model.nncf.do_dummy_forward(force_eval=True)

            for i, observer in enumerate(observers):
                perturbations.add(
                    layer_id=traces_order.get_execution_index_by_traces_index(i),
                    qconfig=qconfig_sequence[i],
                    perturbation=observer.get_observation().to(self._init_device),
                )

            for handle in hook_handles:
                handle.remove()
            observers_for_all_qconfig_sequences.append(observers)

        return perturbations, observers_for_all_qconfig_sequences

    @staticmethod
    def calc_hawq_metric_per_qconfig_sequence(
        qconfig_sequences_in_trace_order: List[QConfigSequenceForHAWQToEvaluate],
        perturbations: Perturbations,
        traces_per_layer: TracesPerLayer,
        device,
    ) -> List[Tensor]:
        metric_per_qconfig_sequence = []
        for qconfig_sequence_in_trace_order in qconfig_sequences_in_trace_order:
            hawq_metric = torch.Tensor([0]).to(device)
            for trace_index, qconfig in enumerate(qconfig_sequence_in_trace_order):
                execution_index = traces_per_layer.traces_order.get_execution_index_by_traces_index(trace_index)
                hawq_metric += traces_per_layer.get_by_trace_index(trace_index) * perturbations.get(
                    layer_id=execution_index, qconfig=qconfig
                )
            metric_per_qconfig_sequence.append(hawq_metric)
        return metric_per_qconfig_sequence

    @staticmethod
    def choose_qconfig_sequence(
        metric_per_qconfig_sequences: List[Tensor], compression_ratio_per_qconfig: List[float], compression_ratio
    ) -> int:
        num_qconfig_sequences = len(metric_per_qconfig_sequences)

        sorted_compression_ratio_per_qconfig = sorted(compression_ratio_per_qconfig)
        indexes_of_sorted_compression_ratio = [
            x[0] for x in sorted(enumerate(compression_ratio_per_qconfig), reverse=False, key=lambda x: x[1])
        ]

        boundary_index = bisect_left(sorted_compression_ratio_per_qconfig, compression_ratio)
        indexes_to_check = [
            indexes_of_sorted_compression_ratio[i] for i in range(boundary_index, num_qconfig_sequences)
        ]
        best_metric = min(list(itemgetter(*indexes_to_check)(metric_per_qconfig_sequences)))
        best_qconfig_sequence_index = metric_per_qconfig_sequences.index(best_metric)
        return best_qconfig_sequence_index

    def get_quantizer_setup_for_qconfig_sequence(
        self, qconfig_sequence_in_traces_order: QConfigSequenceForHAWQToEvaluate, traces_order: TracesOrder
    ) -> SingleConfigQuantizerSetup:
        wqp_ids_in_trace_order = self._get_weight_qp_ids_in_trace_order(traces_order)
        ctrl = self._algo

        quantizer_setup_to_set = self._apply_qconfig_sequence_to_quantizer_setup(
            qconfig_sequence_in_traces_order, wqp_ids_in_trace_order, ctrl.get_quantizer_setup_for_current_state()
        )

        assert quantizer_setup_to_set.shared_input_operation_set_groups
        for group in quantizer_setup_to_set.shared_input_operation_set_groups.values():
            weight_qp_ids = []
            act_qp_ids = []
            for qp_id in group:
                qp = quantizer_setup_to_set.quantization_points[qp_id]
                if qp.is_weight_quantization_point():
                    weight_qp_ids.append(qp_id)
                elif qp.is_activation_quantization_point():
                    act_qp_ids.append(qp_id)
            weight_qps = [quantizer_setup_to_set.quantization_points[qp_id] for qp_id in weight_qp_ids]
            weight_bitwidth_set = {weight_qp.qconfig.num_bits for weight_qp in weight_qps}

            if self._bitwidth_assignment_mode == BitwidthAssignmentMode.STRICT:
                quantizer_setup_to_set = self._set_activations_bitwidth_strictly(
                    quantizer_setup_to_set, act_qp_ids, weight_bitwidth_set
                )
            else:
                quantizer_setup_to_set = self._set_activation_bitwidth_liberally(
                    quantizer_setup_to_set, act_qp_ids, weight_bitwidth_set
                )

        return quantizer_setup_to_set

    def _set_activation_bitwidth_liberally(
        self,
        quantizer_setup_to_set: SingleConfigQuantizerSetup,
        act_qp_ids: List[QuantizationPointId],
        weight_bitwidth_set: Set[int],
    ) -> SingleConfigQuantizerSetup:
        for act_qp_id in act_qp_ids:
            original_quant_module_id = self._original_qp_id_vs_quantizer_module_id_dict[act_qp_id]
            activation_bitwidths_vs_qconfig_sequence = self._hw_precision_constraints.get_bitwidth_vs_qconfigs_dict(
                original_quant_module_id
            )
            activation_bitwidth_set = set(activation_bitwidths_vs_qconfig_sequence.keys())
            intersection = activation_bitwidth_set.intersection(weight_bitwidth_set)
            target_qp = quantizer_setup_to_set.quantization_points[act_qp_id]
            if activation_bitwidth_set.__len__() == 1:
                target_bitwidth = activation_bitwidth_set.pop()
            elif intersection:
                target_bitwidth = min(intersection)
            elif activation_bitwidth_set:
                target_bitwidth = min(activation_bitwidth_set)
            elif weight_bitwidth_set:
                target_bitwidth = min(weight_bitwidth_set)
            else:
                continue

            if activation_bitwidths_vs_qconfig_sequence:
                target_qp.qconfig = deepcopy(activation_bitwidths_vs_qconfig_sequence[target_bitwidth][0])
            else:
                # The activation has no constraints, so the config in the setup was defaulted
                # and we can simply adjust the bitwidth
                target_qp.qconfig.num_bits = target_bitwidth

        return quantizer_setup_to_set

    def _set_activations_bitwidth_strictly(
        self,
        quantizer_setup_to_set: SingleConfigQuantizerSetup,
        act_qp_ids: List[QuantizationPointId],
        weight_bitwidth_set: Set[int],
    ) -> SingleConfigQuantizerSetup:
        if len(weight_bitwidth_set) > 1:
            raise nncf.InternalError("Invalid grouping of weight quantizers")
        all_constraints = set()
        original_quant_module_ids = [
            self._original_qp_id_vs_quantizer_module_id_dict[act_qp_id] for act_qp_id in act_qp_ids
        ]
        for act_quant_module_id in original_quant_module_ids:
            all_constraints.update(self._hw_precision_constraints.get_all_unique_bitwidths(act_quant_module_id))
        common_constraints = set(all_constraints)
        for act_quant_module_id in original_quant_module_ids:
            constraint = self._hw_precision_constraints.get_all_unique_bitwidths(act_quant_module_id)
            common_constraints = common_constraints.intersection(constraint)
        if weight_bitwidth_set:
            common_constraints = common_constraints.intersection(weight_bitwidth_set)
        if not common_constraints:
            raise nncf.InternalError("No hardware compatible bitwidth for activation quantizers")
        for act_qp_id in act_qp_ids:
            quant_id = self._original_qp_id_vs_quantizer_module_id_dict[act_qp_id]
            target_bitwidth = sorted(list(common_constraints))[0]
            bitwidths_vs_qconfig_sequence = self._hw_precision_constraints.get_bitwidth_vs_qconfigs_dict(quant_id)
            qconfig_to_select = bitwidths_vs_qconfig_sequence[target_bitwidth][0]
            quantizer_setup_to_set.quantization_points[act_qp_id].qconfig = qconfig_to_select

        return quantizer_setup_to_set

    @staticmethod
    def _filter_qconfig_sequences_by_grouped_weight_quantizers(
        trace_ordered_qconfig_sequences: List[QConfigSequenceForHAWQToEvaluate],
        weight_quantization_ids_by_execution_order: List[QuantizerId],
        groups_of_adjacent_quantizers: GroupsOfAdjacentQuantizers,
        traces_order: TracesOrder,
    ) -> List[QConfigSequenceForHAWQToEvaluate]:
        """
        Removes configs where adjacent weight quantizers have different bitwidth. Adjacency is defined by common
        activation quantizers
        """
        filtered_qconfig_sequences = []
        all_grouped_indexes = []
        for group_of_adjacent_quantizers in groups_of_adjacent_quantizers:
            wqs = group_of_adjacent_quantizers.weight_quantizers
            if len(wqs) > 1:
                indexes_of_grouped_wq = []
                for quantizer_id, _ in wqs:
                    if quantizer_id in weight_quantization_ids_by_execution_order:
                        index_by_execution_order = weight_quantization_ids_by_execution_order.index(quantizer_id)
                        indexes_of_grouped_wq.append(index_by_execution_order)
                all_grouped_indexes.append(indexes_of_grouped_wq)

        if not all_grouped_indexes:
            return trace_ordered_qconfig_sequences

        for qconfig_sequence in trace_ordered_qconfig_sequences:
            execution_ordered_qconfig_sequence = traces_order.get_execution_order_configs(qconfig_sequence)
            bitwidth_sequence = [qconfig.num_bits for qconfig in execution_ordered_qconfig_sequence]
            keep_config = True
            for indexes_of_grouped_wq in all_grouped_indexes:
                grouped_bits = [bitwidth_sequence[index] for index in indexes_of_grouped_wq]
                if grouped_bits[1:] != grouped_bits[:-1]:
                    keep_config = False
                    break
            if keep_config:
                filtered_qconfig_sequences.append(qconfig_sequence)

        return filtered_qconfig_sequences

    def _filter_qconfig_sequences_by_excessive_bitwidth(
        self, weight_qconfig_sequences_in_trace_order: List[QConfigSequenceForHAWQToEvaluate]
    ) -> List[QConfigSequenceForHAWQToEvaluate]:
        result = weight_qconfig_sequences_in_trace_order
        if self._hw_precision_constraints:
            all_weight_bitwidths = set()
            for wq_id in self._algo.weight_quantizers:
                all_weight_bitwidths.update(self._hw_precision_constraints.get_all_unique_bitwidths(wq_id))

            all_activation_bitwidths = set()
            for aq_id in self._algo.non_weight_quantizers:
                all_activation_bitwidths.update(self._hw_precision_constraints.get_all_unique_bitwidths(aq_id))

            excessive_weight_bitwidths = all_weight_bitwidths - all_activation_bitwidths

            def filter_fn(qconfig_sequence: QConfigSequenceForHAWQToEvaluate):
                all_qconfig_bitwidths = set(map(lambda qconfig: qconfig.num_bits, qconfig_sequence))
                return any(map(lambda x: x not in all_qconfig_bitwidths, excessive_weight_bitwidths))

            if excessive_weight_bitwidths:
                result = list(filter(filter_fn, weight_qconfig_sequences_in_trace_order))
        return result



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from nncf.torch.dynamic_graph.patch_pytorch import register_operator
from nncf.torch.functions import STThreshold
from nncf.torch.functions import logit


def binary_mask(mask):
    return STThreshold.apply(torch.sigmoid(mask))


@register_operator()
def calc_rb_binary_mask(mask, uniform_buffer, eps):
    if uniform_buffer is not None:
        uniform_buffer.uniform_()
        mask = mask + logit(uniform_buffer.clamp(eps, 1 - eps))
    return binary_mask(mask)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from typing import Any, Dict, List, Optional, TypeVar

from nncf import Dataset
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.logging.track_progress import track
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.experimental.tensor import TensorDataType
from nncf.experimental.tensor import functions as fns
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionParameters
from nncf.quantization.algorithms.weight_compression.weight_lowering import do_dequantization
from nncf.quantization.algorithms.weight_compression.weight_lowering import do_integer_quantization
from nncf.quantization.algorithms.weight_compression.weight_lowering import reshape_weight_for_grouped_quantization

TModel = TypeVar("TModel")
TTensor = TypeVar("TTensor")
TWeightType = TypeVar("TWeightType")


class ScaleEstimation:
    """
    Scale estimation algorithm implementation.
    """

    def __init__(
        self,
        model: TModel,
        name_to_node_mapping: Dict[str, Any],
        all_weight_params: List[WeightCompressionParameters],
        nodes_to_compress: List[NNCFNode],
        activations: Optional[Dict[str, TTensor]] = None,
        subset_size: int = 32,
        initial_steps: int = 5,
        scale_steps: int = 10,
        weight_penalty: float = -1.0,
    ):
        """
        :param model: Model for applying algorithm.
        :param name_to_node_mapping: Name to node mapping for updating node weights.
        :param all_weight_params: List of all weight parameters.
        :param nodes_to_compress: List of nodes for processing.
        :param activations: The input activations of the layers considered for compression.
        :param subset_size: The number of samples for scale estimation.
        :param initial_steps: The number of the steps for absmax scale rectification.
        :param scale_steps: The number of the steps for grid search scale rectification
                            from 1.0 to 1.0 - 0.05 * scale_step.
        :param weight_penalty: coefficient for penalty between fp and compressed weights. If -1 then doesn't apply.
        """
        super().__init__()
        self.name_to_node_mapping = name_to_node_mapping
        self._all_weight_params = all_weight_params
        self._nodes_to_compress = nodes_to_compress
        self._activations = activations
        self._subset_size = subset_size
        self._initial_steps = initial_steps
        self._scale_steps = scale_steps
        self._weight_penalty = weight_penalty

        self._set_backend_entity(model)

    @property
    def available_backends(self) -> List[BackendType]:
        return [BackendType.OPENVINO]

    def _set_backend_entity(self, model: TModel) -> None:
        """
        Creates a helper class with a backed-specific logic of the algorithm.

        :param model: Backend-specific input model.
        :param all_weight_params: List of all weight parameters.
        :param nodes_to_compress: List of nodes for processing.
        :param activations: The input activations of the layers considered for compression.
        """

        model_backend = get_backend(model)
        if model_backend == BackendType.OPENVINO:
            from nncf.quantization.algorithms.weight_compression.openvino_backend import OVWeightCompressionAlgoBackend

            self._backend_entity = OVWeightCompressionAlgoBackend(model, self.name_to_node_mapping)
        else:
            raise RuntimeError(
                "Cannot return backend-specific AWQ entity because {} is not supported!".format(model_backend.value)
            )

    def apply(
        self,
        model: TModel,
        graph: NNCFGraph,
        statistic_points: Optional[StatisticPointsContainer] = None,
        dataset: Optional[Dataset] = None,
    ) -> Dict[str, TTensor]:
        """
        Estimates better scale for the int4 nodes in the model.
        Minimizes per-group difference between floating point MatMul and
        MatMul with compressed weights.
        The algorithm computes weighted scale for the group of weights in MatMul, which
        shared the same scale.

        :param model: Model for applying algorithm.
        :param graph: Model graph.
        :param statistic_points: Statistic points with collected statistics values.
        :param dataset: A representative dataset for the calibration process.
        :return: Dict with pairs (weight name, estimated scale).
        """

        compress_decompress_cashe = {}
        res = dict()

        for wp in track(self._all_weight_params, description="Applying Scale Estimation"):
            weight_name = wp.weight_name
            node_name = wp.node_with_weight.node_name
            config = wp.compression_config

            if config.num_bits != 4 or node_name not in self._activations:
                res[weight_name] = None
                continue

            stats = self._activations[node_name]
            reduction_axis = wp.reduction_axes[0]

            cur_config = deepcopy(config)
            cur_config.group_size = -1

            weight_data = self._backend_entity.get_weight_names_and_port_ids(wp.node_with_weight, graph)
            if len(weight_data) != 1:  # not supported by the algorithm
                continue
            _, weight_port_id = weight_data[0]

            X = fns.stack([fns.mean(stat, axis=0) for stat in stats])
            X_full = fns.transpose(X)

            # prevent high memory and time consumption
            if X_full.shape[1] > self._subset_size:
                lens = [stat.shape[0] for stat in stats]
                step = X_full.shape[1] // self._subset_size
                idxs = [i[0] for i in sorted(enumerate(lens), key=lambda x: -x[1])][::step]
                X = X_full[:, idxs]
            else:
                X = X_full

            s = fns.max(fns.abs(X_full), axis=1)

            weight = self._backend_entity.get_weight(wp.node_with_weight, weight_port_id, model, graph)
            weight = weight.astype(TensorDataType.float32)
            eps = fns.finfo(weight).eps

            if reduction_axis == 0:
                weight = fns.transpose(weight)
                reduction_axis = 1

            original_weight = fns.zeros_like(weight) + weight

            compressed_weights, scale, zp = do_integer_quantization(original_weight, reduction_axis, config)
            zp = zp.astype(scale.dtype)

            q_weights = do_dequantization(compressed_weights, scale, zp, reduction_axis)

            s = fns.unsqueeze(s, 0)
            s, _ = reshape_weight_for_grouped_quantization(s, reduction_axis, config.group_size)

            original_weight, _ = reshape_weight_for_grouped_quantization(
                original_weight, reduction_axis, config.group_size
            )

            # all weight in group has importance based on corresponding input activations
            importance = fns.ones_like(original_weight)
            importance = importance * s

            target = compressed_weights.astype(dtype=zp.dtype) - zp
            zero_mask = compressed_weights == zp

            importance = fns.where(zero_mask, 0.0, importance)

            # normalize importances for every group of weights to make sum of them equal to 1.0
            denum = fns.sum(importance, axis=2, keepdims=True)
            importance = importance / (denum + eps)

            X, _ = reshape_weight_for_grouped_quantization(X, 0, config.group_size)
            q_weights, _ = reshape_weight_for_grouped_quantization(q_weights, reduction_axis, config.group_size)
            best_diffs = None
            result_scale = None

            fp_outs = fns.matmul(fns.transpose(original_weight, (1, 0, 2)), X)
            q_outs = fns.matmul(fns.transpose(q_weights, (1, 0, 2)), X)

            # metric for minimization with shape [C_OUT, N_GROUPS], N_GROUPS = C_IN / GROUP_SIZE
            min_max_scale_diffs = fns.mean((fp_outs - q_outs) ** 2, axis=-1)
            min_max_scale_diffs = fns.transpose(min_max_scale_diffs, (1, 0))
            if self._weight_penalty > 0.0:
                min_max_scale_diffs += self._weight_penalty * fns.mean((q_weights - original_weight) ** 2, axis=-1)

            key = (
                (wp.compression_config.mode, wp.compression_config.num_bits) + q_weights.shape + scale.shape + zp.shape
            )
            if key in compress_decompress_cashe:
                compress_decompress_model = compress_decompress_cashe[key]["compress_decompress_model"]
                compress_model = compress_decompress_cashe[key]["compress_model"]
            else:
                compress_decompress_model = self._backend_entity.get_compress_decompress_pipeline(
                    wp, q_weights.shape, scale.shape, zp.shape
                )
                compress_model = self._backend_entity.get_compress_pipeline(wp, q_weights.shape, scale.shape, zp.shape)
                compress_decompress_cashe[key] = {
                    "compress_decompress_model": compress_decompress_model,
                    "compress_model": compress_model,
                }

            zero_scale = 0.001
            zero_mask = zero_scale * zero_mask.astype(original_weight.dtype)

            # iterative rectification of initial scale
            for i in range(self._initial_steps):
                ideal_scale = fns.abs(original_weight) / (fns.abs(target) + zero_mask)
                weighted_scale = ideal_scale * importance

                near_to_ideal_scale = fns.sum(weighted_scale, axis=2, keepdims=True)

                out = compress_decompress_model(original_weight.data, near_to_ideal_scale.data, zp.data)
                q_weights_ = fns.zeros_like(original_weight) + out
                q_outs = fns.matmul(fns.transpose(q_weights_, (1, 0, 2)), X)

                ideal_scale_diffs = fns.mean((fp_outs - q_outs) ** 2, axis=-1)
                ideal_scale_diffs = fns.transpose(ideal_scale_diffs, (1, 0))
                if self._weight_penalty > 0.0:
                    ideal_scale_diffs += self._weight_penalty * fns.mean((q_weights_ - original_weight) ** 2, axis=-1)

                if best_diffs is None:
                    best_diffs = min_max_scale_diffs

                mask = (ideal_scale_diffs > best_diffs).astype(best_diffs.dtype)

                best_diffs = mask * best_diffs + (1.0 - mask) * ideal_scale_diffs

                mask = fns.unsqueeze(mask, axis=2)

                if result_scale is None:
                    near_to_ideal_scale = mask * scale + (1.0 - mask) * near_to_ideal_scale
                else:
                    near_to_ideal_scale = mask * result_scale + (1.0 - mask) * near_to_ideal_scale
                result_scale = near_to_ideal_scale

                if i < self._initial_steps - 1:
                    out = compress_model(original_weight.data, near_to_ideal_scale.data, zp.data)
                    compressed_weights = fns.zeros_like(original_weight) + out
                    target = compressed_weights - zp
                    zero_mask = compressed_weights == zp
                    zero_mask = zero_scale * zero_mask.astype(original_weight.dtype)

            # iterative rectification of scale based on grid search
            for scale_steps in range(self._scale_steps):
                factor = 1.0 - 0.05 * scale_steps
                scaled_scale = factor * scale

                out = compress_model(original_weight.data, scaled_scale.data, zp.data)
                compressed_weights = fns.zeros_like(original_weight) + out

                target = compressed_weights - zp
                zero_mask = compressed_weights == zp
                zero_mask = zero_scale * zero_mask.astype(original_weight.dtype)

                ideal_scale = fns.abs(original_weight) / (fns.abs(target) + zero_mask)
                weighted_scale = ideal_scale * importance
                near_to_ideal_scale = fns.sum(weighted_scale, axis=2, keepdims=True)

                out = compress_decompress_model(original_weight.data, near_to_ideal_scale.data, zp.data)
                q_weights_ = fns.zeros_like(original_weight) + out

                q_outs = fns.matmul(fns.transpose(q_weights_, (1, 0, 2)), X)
                ideal_scale_diffs = fns.mean((fp_outs - q_outs) ** 2, axis=-1)
                ideal_scale_diffs = fns.transpose(ideal_scale_diffs, (1, 0))
                if self._weight_penalty > 0.0:
                    ideal_scale_diffs += self._weight_penalty * fns.mean((q_weights_ - original_weight) ** 2, axis=-1)

                mask = (ideal_scale_diffs > best_diffs).astype(best_diffs.dtype)

                best_diffs = mask * best_diffs + (1.0 - mask) * ideal_scale_diffs

                mask = fns.unsqueeze(mask, axis=2)

                if result_scale is None:
                    near_to_ideal_scale = mask * scale + (1.0 - mask) * near_to_ideal_scale
                else:
                    near_to_ideal_scale = mask * result_scale + (1.0 - mask) * near_to_ideal_scale
                result_scale = near_to_ideal_scale

            res[weight_name] = result_scale

        return res



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional

from nncf.api.statistics import Statistics
from nncf.common.utils.api_marker import api
from nncf.common.utils.helpers import create_table


def _proportion_str(num: int, total_count: int):
    percentage = 100 * (num / max(total_count, 1))
    return f"{percentage:.2f} % ({num} / {total_count})"


class QuantizersCounter:
    def __init__(
        self,
        num_symmetric: int = 0,
        num_asymmetric: int = 0,
        num_signed: int = 0,
        num_unsigned: int = 0,
        num_per_tensor: int = 0,
        num_per_channel: int = 0,
        total_count: int = 0,
        potential_count: Optional[int] = None,
    ):
        """
        Initializes quantizers counter.

        :param num_symmetric: Number of symmetric quantizers.
        :param num_asymmetric: Number of asymmetric quantizers.
        :param num_signed: Number of signed quantizers.
        :param num_unsigned: Number of unsigned quantizers.
        :param num_per_tensor: Number of per-tensor quantizers.
        :param num_per_channel: Number of per-channel quantizers.
        :param total_count: Total count of quantizers.
        :param potential_count: Count of potential quantizers.
        """
        self.num_symmetric = num_symmetric
        self.num_asymmetric = num_asymmetric
        self.num_signed = num_signed
        self.num_unsigned = num_unsigned
        self.num_per_tensor = num_per_tensor
        self.num_per_channel = num_per_channel
        self.total_count = total_count
        self.potential_count = potential_count


def _quantizers_counter_to_rows(counter: QuantizersCounter, qt: str) -> List[List[str]]:
    """
    Converts the counter of quantizers to rows.

    :param counter: Counter of quantizers.
    :param qt: Type of the counter. Takes one of the following values:
        - `WQ` - for the counter of the quantizers for weight.
        - `AQ` - for the counter of the quantizers for activation.
    :return: List of rows.
    """
    rows = [
        [
            f"Symmetric {qt}s / All placed {qt}s",
            _proportion_str(counter.num_symmetric, counter.total_count),
        ],
        [
            f"Asymmetric {qt}s / All placed {qt}s",
            _proportion_str(counter.num_asymmetric, counter.total_count),
        ],
        [
            f"Signed {qt}s / All placed {qt}s",
            _proportion_str(counter.num_signed, counter.total_count),
        ],
        [
            f"Unsigned {qt}s / All placed {qt}s",
            _proportion_str(counter.num_unsigned, counter.total_count),
        ],
        [
            f"Per-tensor {qt}s / All placed {qt}s",
            _proportion_str(counter.num_per_tensor, counter.total_count),
        ],
        [
            f"Per-channel {qt}s / All placed {qt}s",
            _proportion_str(counter.num_per_channel, counter.total_count),
        ],
    ]

    if counter.potential_count:
        rows.append([f"Placed {qt}s / Potential {qt}s", _proportion_str(counter.total_count, counter.potential_count)])

    return rows


@api()
class QuantizationStatistics(Statistics):
    """
    Contains statistics of the quantization algorithm. These statistics include:

    * Information about the share of the quantization, such as:

      * Percentage of symmetric/asymmetric/per-channel/per-tensor weight quantizers relative to the number of placed
        weight quantizers.
      * Percentage of symmetric/asymmetric/per-channel/per-tensor non-weight quantizers relative to the number of
        placed non weight quantizers.
      * Percentage of weight quantizers and non-weight quantizers for each precision relative to the number
        of potential quantizers/placed quantizers.

    * Information about the distribution of the bitwidth of the quantizers.
    * Ratio of enabled quantization.

    .. note:: The maximum possible number of potential quantizers depends on the presence of ignored scopes and the
      mode of quantizer setup that is used at the time of collecting the metric.

    :param wq_counter: Weight quantizers counter.
    :param aq_counter: Activation quantizers counter.
    :param num_wq_per_bitwidth: Number of weight quantizers per bit width.
    :param num_aq_per_bitwidth: Number of activation quantizers per bit width.
    :param ratio_of_enabled_quantizations: Ratio of enabled quantizations.
    """

    def __init__(
        self,
        wq_counter: QuantizersCounter,
        aq_counter: QuantizersCounter,
        num_wq_per_bitwidth: Dict[int, int],
        num_aq_per_bitwidth: Dict[int, int],
        ratio_of_enabled_quantizations: float,
    ):
        self.wq_counter = wq_counter
        self.aq_counter = aq_counter
        self.num_wq_per_bitwidth = num_wq_per_bitwidth
        self.num_aq_per_bitwidth = num_aq_per_bitwidth
        self.ratio_of_enabled_quantizations = ratio_of_enabled_quantizations

    def to_str(self) -> str:
        pretty_strings = []

        table = create_table(
            header=["Statistic's name", "Value"],
            rows=[["Ratio of enabled quantizations", self.ratio_of_enabled_quantizations]],
        )

        pretty_strings.append(f"Statistics of the quantization algorithm:\n{table}")
        pretty_strings.append(self._get_quantization_share_str())
        pretty_strings.append(self._get_bitwidth_distribution_str())
        pretty_string = "\n\n".join(pretty_strings)
        return pretty_string

    def _get_quantization_share_str(self) -> str:
        header = ["Statistic's name", "Value"]

        rows = []
        rows.extend(_quantizers_counter_to_rows(self.wq_counter, "WQ"))
        rows.extend(_quantizers_counter_to_rows(self.aq_counter, "AQ"))

        table = create_table(header, rows)
        pretty_string = f"Statistics of the quantization share:\n{table}"
        return pretty_string

    def _get_bitwidth_distribution_str(self) -> str:
        wq_total_num = sum(self.num_wq_per_bitwidth.values())
        aq_total_num = sum(self.num_aq_per_bitwidth.values())
        q_total_num = wq_total_num + aq_total_num

        bitwidths = self.num_wq_per_bitwidth.keys() | self.num_aq_per_bitwidth.keys()  # union of all bitwidths
        bitwidths = sorted(bitwidths, reverse=True)

        # Table creation
        header = ["Num bits (N)", "N-bits WQs / Placed WQs", "N-bits AQs / Placed AQs", "N-bits Qs / Placed Qs"]
        rows = []
        for bitwidth in bitwidths:
            wq_num = self.num_wq_per_bitwidth.get(bitwidth, 0)  # for current bitwidth
            aq_num = self.num_aq_per_bitwidth.get(bitwidth, 0)  # for current bitwidth
            q_num = wq_num + aq_num  # for current bitwidth

            rows.append(
                [
                    bitwidth,
                    _proportion_str(wq_num, wq_total_num),
                    _proportion_str(aq_num, aq_total_num),
                    _proportion_str(q_num, q_total_num),
                ]
            )

        table = create_table(header, rows)
        pretty_string = f"Statistics of the bitwidth distribution:\n{table}"
        return pretty_string



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.common.graph.patterns.patterns import GraphPattern
from nncf.common.graph.patterns.patterns import IgnoredPatternNames
from nncf.common.utils.registry import Registry
from nncf.openvino.graph.metatypes import openvino_metatypes as om
from nncf.openvino.graph.metatypes.groups import LINEAR_OPERATIONS

OPENVINO_IGNORED_PATTERNS = Registry("IGNORED_PATTERNS")


def _add_softmax_matmul(pattern: GraphPattern) -> None:
    #       SOFTMAX  RESHAPE||TRANSPOSE||GATHER||SQUEEZE||CONCAT
    #           \              /
    #            \            /
    #             \          /
    #              \        /
    #               \      /
    #                MATMUL
    branch_matmul_nodes = [
        om.OVReshapeMetatype,
        om.OVTransposeMetatype,
        om.OVGatherMetatype,
        om.OVSqueezeMetatype,
        om.OVConcatMetatype,
    ]
    softmax = pattern.add_node(**{GraphPattern.LABEL_ATTR: "SOFTMAX", GraphPattern.METATYPE_ATTR: om.OVSoftmaxMetatype})
    matmul = pattern.add_node(**{GraphPattern.LABEL_ATTR: "MATMUL", GraphPattern.METATYPE_ATTR: om.OVMatMulMetatype})
    matmul_branch_nodes = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "RESHAPE||TRANSPOSE||GATHER||SQUEEZE||CONCAT",
            GraphPattern.METATYPE_ATTR: branch_matmul_nodes,
        }
    )
    pattern.add_edge(softmax, matmul)
    pattern.add_edge(matmul_branch_nodes, matmul)


def _add_softmax_reshape_matmul(pattern: GraphPattern) -> None:
    #       SOFTMAX
    #           \
    #            \
    #             \
    #             RESHAPE   RESHAPE||TRANSPOSE||GATHER||SQUEEZE||CONCAT
    #                 \                 /
    #                  \               /
    #                   \             /
    #                    \           /
    #                     \         /
    #                      \       /
    #                        MATMUL
    branch_matmul_nodes = [
        om.OVReshapeMetatype,
        om.OVTransposeMetatype,
        om.OVGatherMetatype,
        om.OVSqueezeMetatype,
        om.OVConcatMetatype,
    ]
    softmax = pattern.add_node(**{GraphPattern.LABEL_ATTR: "SOFTMAX", GraphPattern.METATYPE_ATTR: om.OVSoftmaxMetatype})
    reshape = pattern.add_node(**{GraphPattern.LABEL_ATTR: "RESHAPE", GraphPattern.METATYPE_ATTR: om.OVReshapeMetatype})
    matmul = pattern.add_node(**{GraphPattern.LABEL_ATTR: "MATMUL", GraphPattern.METATYPE_ATTR: om.OVMatMulMetatype})
    matmul_branch_nodes = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "RESHAPE||TRANSPOSE||GATHER||SQUEEZE||CONCAT",
            GraphPattern.METATYPE_ATTR: branch_matmul_nodes,
        }
    )
    pattern.add_edge(softmax, reshape)
    pattern.add_edge(reshape, matmul)
    pattern.add_edge(matmul_branch_nodes, matmul)


@OPENVINO_IGNORED_PATTERNS.register(IgnoredPatternNames.MULTIHEAD_ATTENTION_OUTPUT)
def create_multihead_attention_output() -> GraphPattern:
    pattern = GraphPattern()
    _add_softmax_matmul(pattern)
    _add_softmax_reshape_matmul(pattern)
    return pattern


@OPENVINO_IGNORED_PATTERNS.register(IgnoredPatternNames.FC_BN_HSWISH_ACTIVATION)
def create_fc_bn_hswish() -> GraphPattern:
    pattern = GraphPattern()
    unsqueeze_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "UNSQUEEZE", GraphPattern.METATYPE_ATTR: om.OVUnsqueezeMetatype}
    )
    multiply_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: om.OVMultiplyMetatype}
    )
    add_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD", GraphPattern.METATYPE_ATTR: om.OVAddMetatype})
    squeeze_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "SQUEEZE", GraphPattern.METATYPE_ATTR: om.OVSqueezeMetatype}
    )

    pattern.add_edge(unsqueeze_node, multiply_node)
    pattern.add_edge(multiply_node, add_node)
    pattern.add_edge(add_node, squeeze_node)
    return pattern


@OPENVINO_IGNORED_PATTERNS.register(IgnoredPatternNames.EQUAL_LOGICALNOT)
def create_equal_logicalnot() -> GraphPattern:
    pattern = GraphPattern()
    equal_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "EQUAL", GraphPattern.METATYPE_ATTR: om.OVEqualMetatype})
    logical_not_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "LOGICAL_NOT", GraphPattern.METATYPE_ATTR: om.OVLogicalNotMetatype}
    )

    pattern.add_edge(equal_node, logical_not_node)
    return pattern


@OPENVINO_IGNORED_PATTERNS.register(IgnoredPatternNames.SE_BLOCK)
def create_se_block() -> GraphPattern:
    pattern = GraphPattern()
    any_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "ANY", GraphPattern.METATYPE_ATTR: GraphPattern.NON_PATTERN_NODE_TYPE}
    )
    reduce_mean_node = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "REDUCE_MEAN",
            GraphPattern.METATYPE_ATTR: om.OVReduceMeanMetatype,
            GraphPattern.PATTERN_NODE_TO_EXCLUDE: True,
        }
    )
    linear_node_1 = pattern.add_node(
        **{GraphPattern.METATYPE_ATTR: LINEAR_OPERATIONS, GraphPattern.LABEL_ATTR: "LINEAR"}
    )
    add_node_1 = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD_BIAS", GraphPattern.METATYPE_ATTR: om.OVAddMetatype})
    activation_node_1 = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "RELU, PRELU, SWISH",
            GraphPattern.METATYPE_ATTR: [om.OVReluMetatype, om.OVPReluMetatype, om.OVSwishMetatype],
        }
    )
    linear_node_2 = pattern.add_node(
        **{GraphPattern.METATYPE_ATTR: LINEAR_OPERATIONS, GraphPattern.LABEL_ATTR: "LINEAR"}
    )
    add_node_2 = pattern.add_node(**{GraphPattern.LABEL_ATTR: "ADD_BIAS", GraphPattern.METATYPE_ATTR: om.OVAddMetatype})
    activation_node_2 = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "SIGMOID",
            GraphPattern.METATYPE_ATTR: [om.OVSigmoidMetatype, om.OVHSigmoidMetatype],
        }
    )
    multiply_node = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "MULTIPLY",
            GraphPattern.METATYPE_ATTR: om.OVMultiplyMetatype,
            GraphPattern.PATTERN_NODE_TO_EXCLUDE: True,
        }
    )

    pattern.add_edge(any_node, reduce_mean_node)
    pattern.add_edge(reduce_mean_node, linear_node_1)
    pattern.add_edge(linear_node_1, add_node_1)
    pattern.add_edge(add_node_1, activation_node_1)
    pattern.add_edge(activation_node_1, linear_node_2)
    pattern.add_edge(linear_node_2, add_node_2)
    pattern.add_edge(add_node_2, activation_node_2)
    pattern.add_edge(activation_node_2, multiply_node)
    pattern.add_edge(any_node, multiply_node)
    return pattern



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import os
import textwrap
from abc import ABC
from abc import abstractmethod
from multiprocessing.context import TimeoutError as MPTimeoutError
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import Callable

import torch
from torch.utils.cpp_extension import _get_build_directory

import nncf
from nncf.common.logging import nncf_logger
from nncf.common.logging.logger import extension_is_loading_info_log
from nncf.common.utils.api_marker import api
from nncf.common.utils.registry import Registry

EXTENSIONS = Registry("extensions")

EXTENSION_LOAD_TIMEOUT_ENV_VAR = "NNCF_EXTENSION_LOAD_TIMEOUT"
DEFAULT_EXTENSION_LOAD_TIMEOUT = 60


class ExtensionsType(enum.Enum):
    CPU = 0
    CUDA = 1


def get_build_directory_for_extension(name: str) -> Path:
    build_dir = Path(_get_build_directory("nncf/" + name, verbose=False)) / torch.__version__
    if not build_dir.exists():
        nncf_logger.debug(f"Creating build directory: {str(build_dir)}")
        build_dir.mkdir(parents=True, exist_ok=True)
    return build_dir


class ExtensionLoader(ABC):
    @classmethod
    @abstractmethod
    def extension_type(cls):
        pass

    @classmethod
    @abstractmethod
    def load(cls):
        pass

    @classmethod
    @abstractmethod
    def name(cls) -> str:
        pass

    @classmethod
    def get_build_dir(cls) -> str:
        return str(get_build_directory_for_extension(cls.name()))


class ExtensionLoaderTimeoutException(Exception):
    """Raised when the extension takes too long to load"""


class ExtensionNamespace:
    """
    Provides lazy loading of the underlying extension, i.e. on the first request of a function from the extension.
    """

    def __init__(self, loader: ExtensionLoader):
        """
        :param loader: The extension loader.
        """
        self._loaded_namespace = None
        self._loader = loader

    def get(self, fn_name: str) -> Callable:
        """
        Returns the callable object corresponding to a function from the extension, loading the extension first if
        necessary.
        :param fn_name: A function name as normally accessible from the object return by torch extension's native
          .load() method.
        :return: A callable object corresponding to the requested function.
        """
        if self._loaded_namespace is None:
            timeout = int(os.environ.get(EXTENSION_LOAD_TIMEOUT_ENV_VAR, DEFAULT_EXTENSION_LOAD_TIMEOUT))
            timeout = timeout if timeout > 0 else None

            with extension_is_loading_info_log(self._loader.name()):
                try:
                    pool = ThreadPool(processes=1)
                    async_result = pool.apply_async(self._loader.load)
                    self._loaded_namespace = async_result.get(timeout=timeout)
                except MPTimeoutError as error:
                    msg = textwrap.dedent(
                        f"""\
                        The extension load function failed to execute within {timeout} seconds.
                        This may be due to leftover lock files from the PyTorch C++ extension build process.
                        If this is the case, running the following command should help:
                            rm -rf {self._loader.get_build_dir()}
                        For a machine with poor performance, you may try increasing the time limit by setting the environment variable:
                            {EXTENSION_LOAD_TIMEOUT_ENV_VAR}=180
                        Or disable timeout by set:
                            {EXTENSION_LOAD_TIMEOUT_ENV_VAR}=0
                        For more information, see FAQ entry at: https://github.com/openvinotoolkit/nncf/blob/develop/docs/FAQ.md#importing-anything-from-nncftorch-hangs
                        """  # noqa: E501
                    )
                    raise ExtensionLoaderTimeoutException(msg) from error

        return getattr(self._loaded_namespace, fn_name)


def _force_build_extensions(ext_type: ExtensionsType):
    for class_type in EXTENSIONS.registry_dict.values():
        if class_type.extension_type() != ext_type:
            continue
        class_type.load()


@api(canonical_alias="nncf.torch.force_build_cpu_extensions")
def force_build_cpu_extensions():
    _force_build_extensions(ExtensionsType.CPU)


@api(canonical_alias="nncf.torch.force_build_cuda_extensions")
def force_build_cuda_extensions():
    _force_build_extensions(ExtensionsType.CUDA)


class CudaNotAvailableStub:
    def __getattr__(self, item):
        raise nncf.InstallationError(
            f"CUDA is not available on this machine. Check that the machine has a GPU and a proper "
            f"driver supporting CUDA {torch.version.cuda} is installed."
        )
